import logging
import os
import re
from pathlib import Path
from sys import platform

from pysamosa.common_types import RetrackerBaseType, L1bSourceType
from pysamosa.data_access import data_vars_s3
from pysamosa.retracker_processor import RetrackerProcessor
from pysamosa.settings_manager import get_default_base_settings, SettingsPreset

is_linux = "linux" in platform
is_slurm = "SLURM_JOB_ID" in os.environ
slurm_abs_procid, slurm_n_total_processes = None, None
if is_slurm:
    if "SLURM_ARRAY_TASK_COUNT" in os.environ:
        slurm_n_array_jobs = int(os.environ["SLURM_ARRAY_TASK_COUNT"])
        slurm_array_job_id = int(os.environ["SLURM_ARRAY_TASK_ID"])

        slurm_n_processes_per_job = (
            int(os.environ["SLURM_NPROCS"]) if "SLURM_NPROCS" in os.environ else 1
        )
        n_process_id = int(os.environ["SLURM_PROCID"])

        slurm_n_total_processes = slurm_n_array_jobs * slurm_n_processes_per_job
        slurm_abs_procid = slurm_array_job_id * slurm_n_processes_per_job + n_process_id
    else:
        slurm_n_total_processes = int(os.environ["SLURM_NPROCS"])
        slurm_abs_procid = int(os.environ["SLURM_PROCID"])

    print(
        f"slurm_abs_procid/slurm_n_total_processes: ({slurm_abs_procid}/{slurm_n_total_processes})"
    )

process_rr_files = True

top_coastal_tracks = [
    "371",
    "478",
    "592",
    "074",
    "463",
    "207",
    "508",
    "250",
    "514",
    "491",
    "741",
    "355",
    "181",
    "285",
    "319",
    "493",
    "530",
    "115",
    "703",
    "226",
    "656",
    "513",
    "084",
    "099",
    "646",
    "142",
    "011",
    "393",
    "695",
    "474",
]
rr_track_nums = f'({"|".join(top_coastal_tracks)})'
# rr_track_nums = f'({"|".join(top_coastal_tracks[:1])})'
# rr_track_nums = f'(?!{"|".join(top_coastal_tracks[:5])})[0-9][0-9][0-9]'

name_dest_path = "coral_paper"

# dest_path_base = Path.home() / 'dss' if is_slurm else Path('/nfs/public_ads/Schlembach')
dest_path_base = Path.home() / "dss" if is_slurm else Path("/nfs/DGFI8/H/work_flo/")
dest_path_base = dest_path_base / "pysamosa_results" / name_dest_path

l1bsrc_type, rbt, preset = (
    L1bSourceType.GPOD,
    RetrackerBaseType.SAMPLUS,
    SettingsPreset.CORALv1,
)

# round robin files config
if rbt == RetrackerBaseType.SAM:
    l1b_base_path = (
        "s3a_sr_1_sra_bs"
        if l1bsrc_type is L1bSourceType.EUM_S3
        else "s3a_sr_1_sra_bs_gpod_sam_wfs"
    )
elif rbt == RetrackerBaseType.SAMPLUS or rbt == RetrackerBaseType.SAMPLUSPLUS:
    l1b_base_path = "s3a_sr_1_sra_bs_gpod_samplus_wfs"
# rr_l1b_src_dir = (Path('/nfs/') if is_linux else Path('U:/')) / 'DGFI145/C/seastate_cci/round-robin/satellite/' / l1b_base_path
rr_l1b_src_dir = (
    (Path("/nfs/") if is_linux else Path("U:/"))
    / "DGFI145/C/seastate_cci/round-robin/satellite/"
    / l1b_base_path
)
# rr_l1b_src_dir = (Path('/nfs/') if is_linux else Path('U:/')) / 'DGFI8/H/work_flo/lfsdata/' / l1b_base_path
# rr_l1b_src_dir = Path('/tmp/lfsdata/') / l1b_base_path
# rr_l1b_src_dir = Path('/lfs/DGFI24/data/schlembach/pysamosa_lfsdata/') / l1b_base_path
if is_slurm:
    rr_l1b_src_dir = Path.home() / "dss" / "seastate_cci_rr_lfsdata/" / l1b_base_path

n_offset = 0
n_inds = 0

# n_procs = None
n_procs = 8

# if commented out, RR files will be processed
# src_path = Path('/nfs/DGFI36/altimetry/seastate_cci/round-robin/satellite/s3a_sr_1_sra_bs/')
# l1b_files = [
#     src_path / 'track_207/S3A_SR_1_SRA_BS_20170911T025353_20170911T034422_20171006T170509_3029_022_103______MAR_O_NT_002.SEN3/measurement_l1bs.nc',
#     src_path / 'track_463/S3A_SR_1_SRA_BS_20170411T022013_20170411T031042_20170506T171306_3029_016_231______MAR_O_NT_002.SEN3/measurement_l1bs.nc',
# ]
# l1b_files = [
#     Path('/lfs/DGFI24/fastdata/schlembach/repos/pysamosa/lfsdata/eumetsat/l1bs/S3A_SR_1_SRA_BS_20171210T145735_20171210T154803_20190705T205005_3028_025_239______MR1_R_NT_004.nc'),
# ]


def get_l1b_src_files(*, nc_dest_path, skip_if_exists=True):
    l1b_files = [
        f
        for f in sorted(Path(rr_l1b_src_dir).rglob("*.nc"))
        if bool(re.search(f"(track_){rr_track_nums}", str(f)))
    ]

    if is_slurm:
        # remove files from list if they already exist
        if skip_if_exists:
            l1b_files = [
                f
                for f in l1b_files
                if not (nc_dest_path / str(f.relative_to(rr_l1b_src_dir))).exists()
            ]

        def split(a, n):
            k, m = divmod(len(a), n)
            return (
                a[i * k + min(i, m) : (i + 1) * k + min(i + 1, m)] for i in range(n)
            )

        # split file list into slurm_n_procs chunks of files, take
        # slurm_procid's chunk
        l1b_files = list(split(l1b_files, slurm_n_total_processes))[slurm_abs_procid]

    return l1b_files[::-1]


if __name__ == "__main__":
    if (
        rbt == RetrackerBaseType.SAMPLUS or rbt == RetrackerBaseType.SAMPLUSPLUS
    ) and l1bsrc_type == L1bSourceType.EUM_S3:
        raise RuntimeError(
            "RetrackerBaseType SAM+/SAM++ and L1bSourceType.EUMETSAT is not compatible. "
        )

    nc_dest_path = dest_path_base / (l1bsrc_type.value + "_" + rbt.value)

    (
        rp_sets,
        retrack_sets,
        fitting_sets,
        wf_sets,
        sensor_sets,
    ) = get_default_base_settings(
        retracker_basetype=rbt, settings_preset=preset, l1b_src_type=l1bsrc_type
    )

    rp_sets.nc_dest_dir = nc_dest_path
    rp_sets.n_offset = n_offset
    rp_sets.n_inds = n_inds
    rp_sets.n_procs = n_procs
    # rp_sets.skip_if_exists = False

    l1b_data_vars = data_vars_s3["l1b"]
    if "l1b_files" not in locals():
        l1b_files = get_l1b_src_files(
            nc_dest_path=rp_sets.nc_dest_dir, skip_if_exists=rp_sets.skip_if_exists
        )

    additional_nc_attrs = {
        "L1B source type": l1bsrc_type.value.upper(),
        "Retracker basetype": rbt.value.upper(),
        "Retracker preset": preset.value.upper(),
    }

    rp = RetrackerProcessor(
        l1b_source=l1b_files,
        l1b_data_vars=l1b_data_vars,
        rp_sets=rp_sets,
        retrack_sets=retrack_sets,
        fitting_sets=fitting_sets,
        wf_sets=wf_sets,
        sensor_sets=sensor_sets,
        nc_attrs_kw=additional_nc_attrs,
        # log_level=logging.DEBUG,  #comment in to show
        # debug messages
    )
    rp.process()

    logging.shutdown()



import logging
from pathlib import Path

from pysamosa.common_types import RetrackerBaseType, L1bSourceType
from pysamosa.data_access import data_vars_cs
from pysamosa.retracker_processor import RetrackerProcessor
from pysamosa.settings_manager import get_default_base_settings, SettingsPreset


if __name__ == "__main__":
    nc_src_base_path = Path(
        "/nfs/DGFI145/C/work_flo/cs2_files_samplus_test/CS2_open_ocean/"
    )
    nc_dest_path = Path("/nfs/DGFI145/C/work_flo/cs2_files_samplus_test/")
    run_name = "samplus_cs_test"

    l1b_src_type = L1bSourceType.EUM_CS
    rbt = RetrackerBaseType.SAMPLUS
    pres = SettingsPreset.NONE
    (
        rp_sets,
        retrack_sets,
        fitting_sets,
        wf_sets,
        sensor_sets,
    ) = get_default_base_settings(
        retracker_basetype=rbt, settings_preset=pres, l1b_src_type=l1b_src_type
    )

    rp_sets.nc_dest_dir = nc_dest_path / run_name
    rp_sets.n_offset = 0
    rp_sets.n_inds = 0
    rp_sets.n_procs = 1
    # rp_sets.skip_if_exists = False

    # select files
    # pat = 'SIR_SAR_1B_20150503T160800'
    # l1b_files = [f for f in nc_src_base_path.rglob('*.nc') if bool(re.match(f'(.*){pat}(.*)', str(f)))]
    l1b_files = (
        open("/nfs/DGFI145/C/work_flo/cs2_files_samplus_test/files_CS2_L1B_D.txt")
        .read()
        .splitlines()
    )
    l1b_files = [Path(f) for f in l1b_files]

    additional_nc_attrs = {
        "L1B source type": l1b_src_type.value.upper(),
        "Retracker basetype": rbt.value.upper(),
        "Retracker preset": pres.value.upper(),
    }

    rp = RetrackerProcessor(
        l1b_source=l1b_files,
        l1b_data_vars=data_vars_cs["l1b"],
        rp_sets=rp_sets,
        retrack_sets=retrack_sets,
        fitting_sets=fitting_sets,
        wf_sets=wf_sets,
        sensor_sets=sensor_sets,
        nc_attrs_kw=additional_nc_attrs,
        # log_level=logging.DEBUG,  #comment in to show
        # debug messages
    )
    rp.process()

    logging.shutdown()



samplusplus_test_inds = [
    510,
    2454,
    4980,
    12400,  # problematic with levmar fitting
    38500,  # normal wf
    40154,
    41023,
    12368,
]

# @pytest.mark.parametrize("n_offset", samplusplus_test_inds)
# def test_basic_retracking(n_offset, eumetsat_sam_l1bs, eumetsat_sam_l2):
#     fid = 's3_0'
#     l1b_data = eumetsat_sam_l1bs(n_offset=n_offset, n_inds=1, file_id=fid)
#     l1b_data_single = get_subset_dataset(l1b_data, ind_offset=0)
#     l2_data = eumetsat_sam_l2(n_offset=n_offset, n_inds=1, file_id=fid)
#     l2_data_single = get_subset_dataset(l2_data, ind_offset=0)
#     model_params = get_model_param_obj_from_l1b_data(l1b_data, ind=0)
#
#     sensor_sets = SENSOR_SETS_DEFAULT_S3
#     simple_logger.set_root_logger()
#
#     rt = RetrackerBaseType.SAMPLUSPLUS
#     wf_sets = WaveformSettings.get_default_src_type(L1bSourceType.GPOD)
#     sr = SamosaRetracker(
#         retracker_basetype=rt,
#         retrack_sets=RetrackerSettings(retracker_basetype=rt),
#         fitting_sets=FittingSettings(
#             retracker_basetype=rt,
#         ),
#         sensor_sets=sensor_sets,
#         wf_sets=wf_sets
#     )
#
#     # start fitting
#     res_fit = sr.fit_wf(l1b_data_single=l1b_data_single, model_params=model_params)
#
#     # assert res_fit['misfit'] < 4
#
#     # RIP
#     ripa = RipAnalyser(l1b_data_single['rip'], sensor_sets=SENSOR_SETS_DEFAULT_S3, model_params=model_params)
#
#     # labelling
#     # fig, axs = plt.subplots(2, 1)
#     fig, axs = plt.subplots(2, 1, constrained_layout=True)
#     plot_retrack_result(l1b_data_single, l2_data_single, res_fit, model_params, sensor_sets=sensor_sets, model_sets=sr.model_sets, ax=axs[0], show_l2_model=False, wf_sets=wf_sets)
#     plot_rip_result(l1b_data_single, ripa, ripa.rip_params, ax=axs[1])
#
#     title = f'{eumetsat_sam_l1bs.get_nc_filename(file_id=fid).name} record#: {n_offset}'
#     fig.suptitle(title[:len(title)//2] + '\n' + title[len(title)//2:], fontsize=8)
#
#     fig.show()
#
#
# @pytest.mark.parametrize("n_offset", samplusplus_test_inds)
# def test_rip_analyser(n_offset, eumetsat_sam_l1bs):
#     fid = 's3_0'
#     l1b_data = eumetsat_sam_l1bs(n_offset=n_offset, n_inds=1, file_id=fid)
#     l1b_data_single = get_subset_dataset(l1b_data, ind_offset=0)
#     # l2_data = _read_dataset_vars_from_ds(nc_filename=l2_file, data_var_names=data_vars_l2, n_offset=n_offset, n_inds=n_inds)
#     # l2_data_single = get_single_result_from_dataset(l2_data, ind=0)
#
#     model_params = get_model_param_obj_from_l1b_data(l1b_data, ind=0)
#
#     ripa = RipAnalyser(l1b_data_single['rip'], sensor_sets=SENSOR_SETS_DEFAULT_S3, model_params=model_params)
#
#     # plotting
#     # waveform
#     fig, axs = plt.subplots(2,1)
#     axs[0].plot(l1b_data_single['wf'])
#
#     # RIP
#     plot_rip_result(l1b_data_single, ripa, ripa.rip_params, ax=axs[1])
#
#     fig.show()
#
#
# @pytest.mark.parametrize("n_offset", samplusplus_test_inds)
# def test_gen_2d_rip(n_offset, eumetsat_sam_l1bs):
#     fid = 's3_0'
#     l1b_data = eumetsat_sam_l1bs(n_offset=n_offset, n_inds=1, file_id=fid)
#     l1b_data_single = get_subset_dataset(l1b_data, ind_offset=0)
#     model_params = get_model_param_obj_from_l1b_data(l1b_data, ind=0)
#
#     ripa = RipAnalyser(l1b_data_single['rip'], sensor_sets=SENSOR_SETS_DEFAULT_S3, model_params=model_params)
#
#     n_eff = 53
#     rip_2d_gaussian = ripa.get_gamma0(n_looks_eff=n_eff, n_gates=512)
#
#     # plots
#     fig, axs = plt.subplots(2, 2)
#     axs = axs.ravel()
#     fontsize_labels = 7
#
#     legend_kwargs_default = {'loc': 'upper left',
#                              **{'prop': {'size': fontsize_labels}, 'labelspacing': 0.50, 'borderaxespad': 0.5}}
#
#     # along-track RIP
#     axs[0].set_title('RIP waveform and fitted gaussian', fontsize=fontsize_labels)
#     axs[0].plot(ripa.doppler_beam_inds, ripa.rip_wf_norm, label='measured RIP_az')
#     axs[0].plot(ripa.doppler_beam_inds, ripa.rip_params.rip_az_fitted, '+', label='fitted RIP_az')
#     axs[0].legend(**legend_kwargs_default)
#     axs[0].axvline(0, color='red', linestyle='dashed', linewidth=0.5)
#     axs[0].axvline(ripa.looks_eff[-1], color='red', linestyle='dashed', linewidth=0.5)
#
#     axs[1].set_title('interpolated RIP_act', fontsize=fontsize_labels)
#     axs[1].plot(ripa.rip_act_eff)
#     axs[1].legend()
#
#     axs[2].set_title(f'RIP_az_oversampled, factor={ripa.oversampling_factor}', fontsize=fontsize_labels)
#     axs[2].plot(ripa.rip_wf_oversampled)
#     axs[2].plot(ripa.rip_az_fitted, linestyle='dashed')
#
#     axs[3].set_title('GAMMA0/2D RIP', fontsize=fontsize_labels)
#     axs[3].imshow(rip_2d_gaussian, cmap='jet', interpolation='nearest', aspect='auto');
#
#     # axs[3].set_title('2D RIP/GAMMA0 (measured RIP)')
#     # axs[3].imshow(rip_2d_exact, cmap='jet', interpolation='nearest', aspect='auto');
#
#     fig.show()



import logging
from concurrent import futures
from enum import Enum
from functools import partial

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from pysamosa import retracker
from pysamosa.common_types import (
    ModelSettings,
    RetrackerBaseType,
    RetrackerSettings,
    FittingSettings,
    WaveformSettings,
    SENSOR_SETS_DEFAULT_S3,
)
from pysamosa.data_access import get_model_param_obj_from_l1b_data
from pysamosa.l1b_simulator import L1bSimulator


class CostFunctionType(Enum):
    RMSE = "rmse"
    SDD = "std"
    MEDIAN_BIAS = "median_bias"


default_cost_functions = [
    CostFunctionType.RMSE,
    CostFunctionType.MEDIAN_BIAS,
]


cost_functions = {
    CostFunctionType.RMSE: lambda x, y: np.sqrt(np.mean((x - y) ** 2)),
    CostFunctionType.SDD: lambda x, y: np.std(x - y),
    CostFunctionType.MEDIAN_BIAS: lambda x, y: np.median(x - y),
}


def montesim_call_wrapper(self, kwargs, swh):
    """Wrapper function for ProcessPoolExecutor.map in combination with functools.partial.
    self, kwargs arguments must come first. swh is the one to iterate over.

    :param self: the object of the MonteCarloSimulator
    :param kwargs: the fixed arguments on the call on __call__
    :param swh: the argument to iterate over.
    :return: the wrapped function object (__call__ function of self)
    """
    return self.__call__(swh, **kwargs)


class MonteCarloSimulator:
    def __init__(
        self,
        rbt: RetrackerBaseType,
        retrack_sets: RetrackerSettings,
        fitting_sets: FittingSettings,
        model_sets: ModelSettings,
        wf_sets_retracker: WaveformSettings,
        wf_sets_model: WaveformSettings = None,
    ):
        self.rbt = rbt
        self.retrack_sets = retrack_sets
        self.fitting_sets = fitting_sets
        self.model_sets = model_sets
        self.wf_sets_retracker = wf_sets_retracker
        self.wf_sets_model = (
            wf_sets_model if wf_sets_model is not None else wf_sets_retracker
        )

        self.sr = retracker.SamosaRetracker(
            retracker_basetype=RetrackerBaseType.SAM,
            retrack_sets=retrack_sets,
            fitting_sets=fitting_sets,
            wf_sets=wf_sets_retracker,
            sensor_sets=SENSOR_SETS_DEFAULT_S3,
        )

    def __call__(
        self,
        swh,
        *,
        n_realisations,
        add_set_name=None,
        add_thermal_speckle_noise=True,
        add_interference=False,
    ):
        add_set_name = [] if add_set_name is None else add_set_name
        add_set_name = (
            add_set_name if isinstance(add_set_name, list) else [add_set_name]
        )

        l1b_sim = L1bSimulator(
            model_sets=self.model_sets,
            wf_sets=self.wf_sets_model,
            swh=swh,
            Pu=1.0,
            add_thermal_speckle_noise=add_thermal_speckle_noise,
            add_interference=add_interference,
        )
        l1b_sim_it = iter(l1b_sim)

        logging.info(
            f"Running montecarlo simulation with SWH={l1b_sim.swh}, epoch={l1b_sim.epoch_ns}ns, Pu_norm={l1b_sim.Pu}"
        )

        fit_results = []
        for i in range(n_realisations):
            l1b_data_single = next(l1b_sim_it)
            model_params = get_model_param_obj_from_l1b_data(l1b_data_single, 0)

            res_fit = self.sr.fit_wf(
                l1b_data_single=l1b_data_single, model_params=model_params
            )

            # append expected values: swh, epoch_ns, Pu
            res_fit = {
                **res_fit,
                **{
                    "swh_expected": l1b_sim.swh,
                    "epoch_ns_expected": l1b_sim.epoch_ns,
                    "Pu_expected": l1b_sim.Pu,
                },
            }

            # append additional settings columns
            # hack because of NameError when eval tries to evaluate self
            symbols = {"self": self}
            res_fit = {
                **res_fit,
                **{
                    col.split(".")[-1]: eval(f"self.{col}", symbols)
                    for col in add_set_name
                },
            }

            fit_results.append(res_fit)

        return pd.DataFrame(fit_results)

    def multi_proc(self, swh, max_workers=None, **kwargs):
        swh = np.asarray(swh)

        if max_workers == 1:
            fit_results_list = []
            for s in swh:
                fit_results_list.append(self.__call__(s, **kwargs))
        else:
            with futures.ProcessPoolExecutor(max_workers=max_workers) as pool:
                # treat kwargs as fixed arguments for monte_sim.__call__ and
                # swh as variable param on each process to parallelise
                fit_results_list = pool.map(
                    partial(montesim_call_wrapper, self, kwargs), swh, chunksize=5
                )

        df = pd.concat(fit_results_list, ignore_index=True)

        return df


def plot_cost_func_vs_swh(
    df,
    *,
    var_type,
    cost_func: CostFunctionType,
    ax=None,
    ax_kwargs=None,
    legend_kwargs=None,
    label_kw=None,
    ticks_kw=None,
    subplot_name=None,
    subplot_name_kw=None,
):
    _ax = ax if ax is not None else plt.gca()
    ax_kwargs = ax_kwargs if ax_kwargs is not None else {}
    default_legend_kwargs = {"fontsize": 5}
    legend_kwargs = (
        {**default_legend_kwargs, **legend_kwargs}
        if legend_kwargs is not None
        else default_legend_kwargs
    )

    default_label_kw = {"fontsize": 5}
    label_kw = (
        {**default_label_kw, **label_kw} if label_kw is not None else default_label_kw
    )

    default_ticks_kw = {"labelsize": 5}
    ticks_kw = (
        {**default_ticks_kw, **ticks_kw} if ticks_kw is not None else default_ticks_kw
    )

    subplot_name_kw = subplot_name_kw if subplot_name_kw is not None else {}

    func = cost_functions[cost_func]
    df_func_vs_swh_exp = df.groupby(by="swh_expected").apply(
        lambda g: func(g[var_type], g[f"{var_type}_expected"])
    )

    _ax = df_func_vs_swh_exp.plot(marker="o", markersize=3, ax=_ax, **ax_kwargs)

    label_unit = {
        "swh": ("SWH", "m"),
        "epoch_ns": ("Epoch", "ns"),
        "Pu": ("Pu", "norm"),
    }

    _ax.set_xlabel(
        f"expected {label_unit[var_type][0]} [{label_unit[var_type][1]}]", **label_kw
    )
    _ax.set_ylabel(
        f'{label_unit[var_type][0]} {cost_func.value.lower().replace("_", " ")} [{label_unit[var_type][1]}]',
        **label_kw,
    )
    _ax.set_xticks(np.arange(0, int(df.swh_expected.values[-1]) + 1, 2))
    _ax.tick_params(**ticks_kw)
    _ax.grid()
    _ax.legend(**legend_kwargs)

    if subplot_name:
        _ax.annotate(subplot_name, **subplot_name_kw)

    return _ax


def plot_swh_epoch_scatter(df, ax=None):
    _ax = ax
    # _ax = ax if ax is not None else plt.gca()

    def swh_epoch_diff(g):
        return pd.Series(
            {
                "swh_diff": g["swh"] - g["swh_expected"],
                "epoch_ns_diff": g["epoch_ns"] - g["epoch_ns_expected"],
            }
        )

    new_df = df.apply(swh_epoch_diff, axis=1)
    _ax = new_df.plot.scatter("swh_diff", "epoch_ns_diff", ax=_ax)
    _ax.set_xlabel("SWH RMSE [m]")
    _ax.set_ylabel("Epoch RMSE [m]")
    _ax.grid()

    return _ax



import xarray as xr
import cftime
from pathlib import Path
import tempfile
import netCDF4

import numpy as np
import pandas as pd
from bokeh.io import show, output_file
from bokeh.models import HoverTool
from bokeh.models.glyphs import Patches
from bokeh.plotting import figure, ColumnDataSource
from bokeh.tile_providers import Vendors

from pysamosa.data_access import _read_dataset_vars_from_ds
from pysamosa.settings import LFSDATA_DIR

# data preparation
ncfile_list = sorted((LFSDATA_DIR / "l1bs").rglob("*.nc"))

nc_filenames = [
    Path.cwd().parent
    / "data"
    / "s6"
    / "l2"
    / "S6A_P4_2__HR_STD__NT_038_018_20211120T051224_20211120T060836_F06.nc",
]

for nc_filename in nc_filenames:
    is_s6 = "S6A" in str(nc_filename)
    is_s3 = "S3" in str(nc_filename)
    grps = [
        g
        for g in list(netCDF4.Dataset(nc_filename).groups)
        if "data_20" in g or "data_140" in g
    ]
    nc_grp = grps[0] + "/ku" if len(grps) > 0 else ""

    nc_basename = (
        nc_filename.stem
        if "measurement" not in str(nc_filename)
        else nc_filename.parent.stem
    )

    if "dart" in str(nc_filename):
        cycle, ppass = int(nc_basename.split("_")[13]), int(nc_basename.split("_")[14])
    elif "_2_" in nc_basename:
        cycle, ppass = int(nc_basename.split("_")[8]), int(nc_basename.split("_")[9])
    elif "_1A_" in nc_basename:
        cycle, ppass = int(nc_basename.split("_")[13]), int(nc_basename.split("_")[14])
    elif "_1B_" in nc_basename:
        cycle, ppass = int(nc_basename.split("_")[13]), int(nc_basename.split("_")[14])
    else:
        cycle, ppass = 0, 0

    # level_str = (str(nc_filename.name).split('_')[2] if 'meas' not in str(nc_filename) else str(nc_filename.parent.name).split('_')[2]).lower()
    # # level_str = (str(nc_filename.name).split('_')[2] if 'meas' not in str(nc_filename) else str(nc_filename.name).split('_')[2]).lower()
    # level_str = level_str.replace('_1_', '_1b_')
    #
    # if level_str == '':
    #     level_str = '1a'
    #
    # is_l1a = f'_1a_' in str(nc_filename).lower()
    # is_l2 = f'_2_' in str(nc_filename).lower()

    with xr.open_dataset(nc_filename, decode_times=False, group=nc_grp) as dstmp:
        if is_s6:
            data_vars = {
                "lat_rad": [k for k in list(dstmp.coords.keys()) if "lat" in k][0],
                "lon_rad": [k for k in list(dstmp.coords.keys()) if "lon" in k][0],
                "time": [k for k in list(dstmp.coords.keys()) if "time" in k][0],
            }
        elif is_s3:
            data_vars = {
                "lat_rad": [
                    k for k in list(dstmp.coords.keys()) if "lat" in k and "sar" in k
                ][0],
                "lon_rad": [
                    k for k in list(dstmp.coords.keys()) if "lon" in k and "sar" in k
                ][0],
                "time": [
                    k for k in list(dstmp.coords.keys()) if "time" in k and "sar" in k
                ][0],
            }

    if "model" in str(nc_filename):
        data_vars["angle_of_approach_to_coast"] = "angle_of_approach_to_coast"

    ds = _read_dataset_vars_from_ds(
        nc_filename=nc_filename, data_var_names=data_vars, group=nc_grp
    )
    # nc_filename = LFSDATA_DIR / 'eumetsat/l1bs/S3A_SR_1_SRA_BS_20171210T145735_20171210T154803_20190705T205005_3028_025_239______MR1_R_NT_004.nc'
    # ds = _read_dataset_vars_from_ds(nc_filename=nc_filename, data_var_names=data_vars_eumetsat['l1b'])

    if isinstance(ds["time"][0], cftime.DatetimeGregorian):
        ds["time"] = ds["time"].astype("datetime64[ns]")

    df = pd.DataFrame(
        {
            k: v
            for k, v in ds.items()
            if k
            in [
                "time",
                "lat_rad",
                "lon_rad",
                "alt_m",
                "dist2coast",
                "distance_to_coast",
                "record_inds",
                "angle_of_approach_to_coast",
            ]
        }
    )

    # reduce dataset
    # df = df[::10]

    df["lat"] = np.degrees(df["lat_rad"])
    df["lon"] = np.degrees(df["lon_rad"])

    # df = df.drop(columns=['lat_rad', 'lon_rad'])

    # Define function to switch from lat/lon to mercator coordinates
    def x_coord(lat, lon):
        r_major = 6378137.000
        x = r_major * np.radians(lon)
        scale = x / lon
        y = (
            180.0
            / np.pi
            * np.log(np.tan(np.pi / 4.0 + lat * (np.pi / 180.0) / 2.0))
            * scale
        )
        return (x, y)

    # Define coord as tuple (lat,long)
    df["coordinates"] = list(zip(df["lat"], df["lon"]))

    # Create mercator column in our df
    df["mercator"] = [x_coord(lat, lon) for lat, lon in df["coordinates"]]

    # Split that column out into two separate columns - mercator_x and
    # mercator_y
    df[["mercator_x", "mercator_y"]] = df["mercator"].apply(pd.Series)

    # calc bearings (angles between points):
    def calc_bearings(lat, lon):
        """Calculate bearings between two vector of ellipsoidal coordinates.
        according to https://towardsdatascience.com/calculating-the-bearing-between-two-geospatial-coordinates-66203f57e4b4
        0.0 -> North, -pi/2 -> East, -pi -> South, -3/2 pi -> West

        returns bearings (in radians)
        """
        _lat = np.radians(np.asarray(lat))
        _lon = np.radians(np.asarray(lon))

        bearings = np.zeros(lat.size)
        for i in range(lat.size):
            if i != (lat.size - 1):
                lat0, lon0 = _lat[i], _lon[i]
                lat1, lon1 = _lat[i + 1], _lon[i + 1]
                dL = lon1 - lon0
                X = np.cos(lat1) * np.sin(dL)
                Y = np.cos(lat0) * np.sin(lat1) - np.sin(lat0) * np.cos(lat1) * np.cos(
                    dL
                )

                bearings[i] = np.arctan2(X, Y)
            else:
                bearings[i] = bearings[i - 1]

        bearings += np.pi  # so that it fits the definition
        return bearings

    bearings = calc_bearings(df["lat"], df["lon"])
    df["act_bearing"] = -bearings

    mercator_scale_factor = 1 / np.cos(df["lat_rad"])
    df["footprint_size_x"] = mercator_scale_factor * 10000
    df["footprint_size_y"] = mercator_scale_factor * 300

    # Plotting
    # palette = PRGn[11]  # Choose palette

    # Define color mapper - which column will define the colour of the data points
    # color_mapper = linear_cmap(field_name = 'AveragePrice', palette = palette, low = df['AveragePrice'].min(), high = df['AveragePrice'].max())

    # Set tooltips - these appear when we hover over a data point in our map,
    # very nifty and very useful
    tooltips = [
        ("latitude", "@lat"),
        ("longitude", "@lon"),
        ("dist2coast", "@dist2coast"),
        ("distance_to_coast", "@distance_to_coast"),
        ("altitude", "@alt_m"),
        ("record_ind", "@record_inds"),
        ("buoy_label", "@buoy_label"),
        ("angle_of_approach_to_coast", "@angle_of_approach_to_coast"),
    ]

    # Create the figure
    p = figure(
        title=f"{nc_basename}",
        # plot_width=1200, plot_height=900,
        sizing_mode="scale_width",
        aspect_ratio=2.0,
        x_axis_type="mercator",
        y_axis_type="mercator",
        x_axis_label="longitude",
        y_axis_label="latitude",
        # tooltips=tooltips,
        tools="pan,wheel_zoom,box_zoom,save",
        toolbar_location="above",
        output_backend="webgl",
    )

    # Add map tile
    p.add_tile(Vendors.ESRI_IMAGERY)

    # Tell Bokeh to use df as the source of the data
    source = ColumnDataSource(data=df)
    # glyph = Rect(x="mercator_x", y="mercator_y", width='footprint_size_x', height='footprint_size_y', angle='act_bearing', fill_color='red', fill_alpha=0.6)
    # p.add_glyph(source, glyph)

    ####
    # hack because the hover area of Rect is shift (it is a bug:
    # https://github.com/bokeh/bokeh/issues/9752)
    rect_points_xy = np.array([-0.5, 0.5])
    xp = df["footprint_size_x"].values[:, np.newaxis] * rect_points_xy[np.newaxis]
    yp = df["footprint_size_y"].values[:, np.newaxis] * rect_points_xy[np.newaxis]

    def rotate(p, origin=(0, 0), angle_rad=0):
        R = np.array(
            [
                [np.cos(angle_rad), -np.sin(angle_rad)],
                [np.sin(angle_rad), np.cos(angle_rad)],
            ]
        )
        o = np.atleast_2d(origin)
        p = np.atleast_2d(p)
        return np.squeeze((R @ (p.T - o.T) + o.T).T)

    xs = np.zeros((xp.shape[0], 4))
    ys = np.zeros((yp.shape[0], 4))
    for i in range(xp.shape[0]):
        x = np.array([xp[i, 0], xp[i, 0], xp[i, 1], xp[i, 1]])
        y = np.array([yp[i, 0], yp[i, 1], yp[i, 1], yp[i, 0]])

        # rotate points
        points_rotated = np.apply_along_axis(
            lambda p_i: rotate(p_i, angle_rad=-bearings[i]),
            axis=0,
            arr=np.vstack([x, y]),
        )
        xs[i, :] = points_rotated[0, :]
        ys[i, :] = points_rotated[1, :]

    xs = df["mercator_x"].values[:, np.newaxis] * np.ones((1, xs.shape[1])) + xs
    ys = df["mercator_y"].values[:, np.newaxis] * np.ones((1, ys.shape[1])) + ys

    source.add(xs.tolist(), "xs")
    source.add(ys.tolist(), "ys")

    glyph = Patches(
        xs="xs",
        ys="ys",
        fill_color="red",
        fill_alpha=0.6,
    )
    ####
    p.add_glyph(source, glyph)

    # Add a hover tool referring to the formatted columns
    hover = HoverTool(tooltips=tooltips)
    hover.point_policy = "follow_mouse"
    # hover.point_policy = 'snap_to_data'
    # hover.anchor = 'top_left'

    # Add the hover tool to the graph
    p.add_tools(hover)

    # Display in notebook
    output_file(
        Path(tempfile.gettempdir()) / "track_browser.html", title=f"{cycle}/{ppass}"
    )

    show(p)



import tempfile
from pathlib import Path
import os

import numpy as np
import xarray as xr
import pandas as pd
from datetime import datetime
from pysamosa.dist2coast import get_dist_pacioos

from .common_types import ModelParameter
from .conf_params import CONST_C

data_vars_s3 = {
    "l1b": {
        "wf": "i2q2_meas_ku_l1b_echo_sar_ku",
        "time": "time_l1b_echo_sar_ku",
        "lat_rad": "lat_l1b_echo_sar_ku",
        "lon_rad": "lon_l1b_echo_sar_ku",
        "alt_m": "alt_l1b_echo_sar_ku",
        "h_rate_m_per_s": "orb_alt_rate_l1b_echo_sar_ku",
        "Vs_m_per_s": "x_vel_l1b_echo_sar_ku",
        # 'ksix_rad': 'pitch_sral_mispointing_l1b_echo_sar_ku',  # pitch
        # 'ksiy_rad': 'roll_sral_mispointing_l1b_echo_sar_ku',  # roll
        "beam_ang_stack_rad": "beam_ang_stack_l1b_echo_sar_ku",
        "epoch_ref_gate": "epoch_ref_gate_const",
        "tracker_range_m": "range_ku_l1b_echo_sar_ku",
    },
    "l2": {
        "time": "time_l1bs_echo_sar_ku",
        "lat_rad": "lat_l1bs_echo_sar_ku",
        "lon_rad": "lon_l1bs_echo_sar_ku",
        "swh": "swh_ocean_20_ku",
        "swh_qual": "swh_ocean_qual_20_ku",
        "Pu_W": "amplitude_ocean_20_ku",
        "epoch_ns": "epoch_ocean_20_ku",
        "range": "range_ocean_20_ku",
        "tracker_range_m": "tracker_range_20_ku",
        "alt_m": "alt_20_ku",
        "n_iter": "number_of_iterations_ocean_20_ku",
        "misfit": "mqe_ocean_20_ku",
        # 'l2norm': 'Misfit_20Hz',
        # 'p_out_w': 'Pout_20Hz',
        # 'mss': 'MSS_20Hz',
    },
}

data_vars_cs = {
    "l1b": {
        "wf": "pwr_waveform_20_ku",
        "time": "time_20_ku",
        "lat_rad": "lat_20_ku",
        "lon_rad": "lon_20_ku",
        "alt_m": "alt_20_ku",
        "h_rate_m_per_s": "orb_alt_rate_20_ku",
        # 'Vs_m_per_s': 'sat_vel_vec_20_ku',
        "Vs_m_per_s": "sat_vel_vec_20_ku",
        "ksix_rad": "off_nadir_pitch_angle_str_20_ku",  # pitch
        "ksiy_rad": "off_nadir_roll_angle_str_20_ku",  # roll
        "beam_ang_stack_rad": "beam_dir_vec_20_ku",
        "epoch_ref_gate": "epoch_ref_gate_const",
        "tracker_range_m": "window_del_20_ku",
    },
    # #GOPR
    # 'l1b': {
    #     'wf': 'pwr_waveform_20_hr_ku',
    #     'time': 'time_20_hr_ku',
    #     'lat_rad': 'lat_20_hr_ku',
    #     'lon_rad': 'lon_20_hr_ku',
    #     'alt_m': 'alt_20_hr_ku',
    #     'h_rate_m_per_s': 'orb_alt_rate_20_hr_ku',
    #     # 'Vs_m_per_s': 'sat_vel_vec_20_hr_ku',
    #     'Vs_m_per_s': 'sat_vel_vec_20_hr_ku',
    #     'ksix_rad': 'off_nadir_pitch_angle_str_20_hr_ku',  # pitch
    #     'ksiy_rad': 'off_nadir_roll_angle_str_20_hr_ku',  # roll
    #     'beam_ang_stack_rad': 'beam_dir_vec_20_hr_ku',
    #     'epoch_ref_gate': 'epoch_ref_gate_const',
    #     'tracker_range_m': 'window_del_20_hr_ku',
    # },
    "l2": {
        "time": "time_20_ku",
        "lat_rad": "lat_20_ku",
        "lon_rad": "lon_20_ku",
        "swh": "swh_ocean_20_ku",
        "swh_qual": "swh_ocean_qual_20_ku",
        "Pu_W": "const",
        # 'epoch_ns': 'const',
        "range": "range_ocean_20_ku",
        # 'tracker_range_m': 'window_del_20_hr_ku',
        "alt_m": "alt_20_ku",
        "n_iter": "const",
        "misfit": "mqe_ocean_20_ku",
        # gpod
        # 'time': 'time_counter_20Hz',
        # 'lat_rad': 'latitude_20Hz',
        # 'lon_rad': 'longitude_20Hz',
        # 'swh': 'SWH_20Hz',
        # 'swh_qual': 'Misfit_20Hz',
        # 'Pu_W': 'Pout_20Hz',
        # 'epoch_ns': 'Epoch_20Hz',
        # 'range': 'Range_Unc_20Hz',
        # 'tracker_range_m': 'Window_Delay_20Hz',
        # 'alt_m': 'altitude_20Hz',
        # 'n_iter': 'Iterations_Count_20Hz',
        # 'misfit': 'Misfit_20Hz',
    },
}

data_vars_gpod = {
    "l1b": {
        "wf": "SAR_Echo_Data",
        "time": "time_counter_20Hz",
        "lat_rad": "latitude_20Hz",
        "lon_rad": "longitude_20Hz",
        "alt_m": "altitude_20Hz",
        "h_rate_m_per_s": "altitude_rate_20Hz",
        "Vs_m_per_s": "satellite_velocity_20Hz",
        "ksix_rad": "pitch_mispointing_20Hz",  # pitch
        "ksiy_rad": "roll_mispointing_20Hz",  # roll
        "beam_ang_stack_rad": "Stack_Boresight_Beam_Angle",
        "epoch_ref_gate": "Epoch_Reference_Gate",
        "tracker_range_m": "Window_Delay_20Hz",
        "rip": "Substack_RIP_Data",
        "n_looks_stack": "NLook_20Hz",
        "oversampling_factor": "const",
        "thermal_noise": "ThN_20Hz",
    },
    "l2": {
        "time": "time_counter_20Hz",
        "lat_rad": "latitude_20Hz",
        "lon_rad": "longitude_20Hz",
        "swh": "SWH_20Hz",
        "swh_qual": "Misfit_20Hz",
        "Pu_W": "Pout_20Hz",
        "epoch_ns": "Epoch_20Hz",
        "range": "Range_Unc_20Hz",
        "tracker_range_m": "Window_Delay_20Hz",
        "alt_m": "altitude_20Hz",
        "n_iter": "Iterations_Count_20Hz",
        "misfit": "Misfit_20Hz",
    },
}


data_vars_s6 = {
    "l1b": {
        "wf": "power_waveform",
        "time": "time",
        "lat_rad": "latitude",
        "lon_rad": "longitude",
        "alt_m": "altitude",
        "h_rate_m_per_s": "altitude_rate",
        "Vs_m_per_s": "velocity_vector",
        "ksix_rad": "off_nadir_pitch_angle_pf",  # pitch
        "ksiy_rad": "off_nadir_roll_angle_pf",  # roll
        "beam_ang_stack_rad": "look_angle_start",
        "stack_mask_start_stop": "stack_mask_start_stop",
        "epoch_ref_gate": "epoch_ref_gate_const",
        "tracker_range_m": "tracker_range_calibrated",
        "pri_hz": "pulse_repetition_interval",
    },
    "l2": {
        "time": "time",
        "lat_rad": "latitude",
        "lon_rad": "longitude",
        "swh": "swh_ocean",
        "swh_qual": "swh_ocean_qual",
        "Pu_W": "amplitude_ocean",
        "epoch_ns": "epoch_ocean",
        "range": "range_ocean",
        "tracker_range_m": "tracker_range_calibrated",
        "alt_m": "altitude",
        "n_iter": "num_iterations_ocean",
        "misfit": "mqe_ocean",
    },
}

data_vars_dart = {
    "l1b": {
        "wf": "power_waveform_pseudo_dda",
        "time": "time",
        "lat_rad": "latitude",
        "lon_rad": "longitude",
        "alt_m": "altitude",
        "h_rate_m_per_s": "altitude_rate",
        "Vs_m_per_s": "velocity_vector",
        "ksix_rad": "off_nadir_pitch_angle_pf",  # pitch
        "ksiy_rad": "off_nadir_roll_angle_pf",  # roll
        "beam_ang_stack_rad": "look_angles",
        "stack_mask_start_stop": "stack_mask_start_stop",
        "epoch_ref_gate": "epoch_ref_gate_var",
        "tracker_range_m": "tracker_range_calibrated",
        "pri_hz": "pulse_repetition_interval",
    },
    "l1b_ffsar": {
        "wf": "power_waveform",
        "time": "time",
        "lat_rad": "latitude",
        "lon_rad": "longitude",
        "alt_m": "altitude",
        "h_rate_m_per_s": "altitude_rate",
        "Vs_m_per_s": "velocity_vector",
        "ksix_rad": "off_nadir_pitch_angle_pf",  # pitch
        "ksiy_rad": "off_nadir_roll_angle_pf",  # roll
        "beam_ang_stack_rad": "look_angles",
        "stack_mask_start_stop": "stack_mask_start_stop",
        "epoch_ref_gate": "epoch_ref_gate_var",
        "tracker_range_m": "tracker_range_calibrated",
        "pri_hz": "pulse_repetition_interval",
    },
    "l2": {
        "time": "time",
        "lat_rad": "latitude",
        "lon_rad": "longitude",
        "swh": "swh_ocean",
        "swh_qual": "swh_ocean_qual",
        # 'Pu_W': 'amplitude_ocean',
        # 'epoch_ns': 'epoch_ocean',
        "range": "range_ocean",
        "tracker_range_m": "tracker_range_calibrated",
        "alt_m": "altitude",
        # 'n_iter': 'num_iterations_ocean',
        # 'misfit': 'mqe_ocean',
    },
}

data_vars_retracker = {
    "l2": {
        "time": "time",
        "lat_rad": "latitude",
        "lon_rad": "longitude",
        "swh": "swh",
        "swh_qual": "swh_qual",
        "Pu_W": "Pu",
        "epoch_ns": "epoch",
        "range": "range",
        "tracker_range_m": "tracker_range",
        "alt_m": "altitude",
        # 'n_iter': '',
        "misfit": "misfit",
    }
}


def get_last_ind(n_offset, n_inds, n_total_length):
    """Returns last (0-based) index (including it) by considering n_offset, n_inds and the maximal ind (end of file)

    :param n_offset: start index
    :param n_inds: number of indices
    :param n_total_length: maximum number of samples
    :return: index number
    """
    if n_inds == 0:
        ind_last = n_total_length - 1
    elif n_offset + n_inds >= n_total_length:
        ind_last = n_total_length - 1
    else:
        ind_last = n_offset + n_inds - 1

    if ind_last > n_total_length:
        raise RuntimeError(
            f"get_last_ind: last_ind ({ind_last}) exceeds total length ({n_total_length})"
        )

    return ind_last


def convert_eum_mqe_to_misfit(mqe):
    return np.sqrt(100 * np.sqrt(mqe))


def _read_dataset_vars_from_ds(
    *,
    nc_filename,
    data_var_names,
    n_offset=0,
    n_inds=0,
    do_interp_dist2coast=False,
    group=None,
):
    # hack required because some nc files have issues with decoding of times
    try:
        ds = xr.open_dataset(nc_filename, group=group)
    except BaseException:
        ds = xr.open_dataset(nc_filename, group=group, decode_times=False)

    last_ind = get_last_ind(
        n_offset=n_offset,
        n_inds=n_inds,
        n_total_length=ds[data_var_names["time"]].shape[0],
    )

    def get_data_var(_v):
        if "const" in _v:
            var = None
        elif ds[_v].size == 1:
            var = float(ds[_v].values)
        else:
            var = ds[_v][n_offset : last_ind + 1].values

        # general conversions
        if (
            "lat" in _v.lower()
            or "lon" in _v.lower()
            or "off_nadir" in _v.lower()
            or "beam" in _v.lower()
        ):
            is_deg_unit = (
                hasattr(ds[_v], "units") and "deg" in str(ds[_v].units).lower()
            )
            # is_rad_unit = (hasattr(ds[_v], 'units') and 'rad' in ds[_v].units)
            if is_deg_unit or (
                not is_deg_unit and any((abs(ds[_v]) > np.pi / 2).values.ravel())
            ):
                var = np.radians(var)

        return var

    dataset = {}
    for k, v in data_var_names.items():
        if v in dir(ds) or "const" in v:
            is_eumetsat = (
                ("contact" in dir(ds) and "eumetsat" in ds.contact)
                or "EUM" in str(nc_filename)
            ) and "processor" not in dir(ds)
            is_gpod = (
                "NetCDF_CreatedBy" in dir(ds) and "SARvatore" in ds.NetCDF_CreatedBy
            )
            is_dart = "title" in dir(ds) and "dart" in ds.title.lower()
            is_s3 = "S3" in str(nc_filename)
            is_cs = "/CS" in str(nc_filename)
            is_cs_gopr = is_cs and "GOPR" in str(nc_filename)
            is_s6 = "S6" in str(nc_filename)

            # provider-specific conversions
            # GPOD
            if is_gpod and "oversampling_factor" in k:
                # zero_pad_factor = 2.0 if hasattr(ds, 'L1b_HISTORY') and 'Zero_Padding_Flag = Y' in ds.L1b_HISTORY else 1.0
                # data_var = zero_pad_factor * wf_oversampling_factor
                wf_oversampling_factor = (
                    float(ds.WAVEFORM_OVERSAMPLING_FACTOR)
                    if hasattr(ds, "WAVEFORM_OVERSAMPLING_FACTOR")
                    else 1
                )
                data_var = wf_oversampling_factor
            elif is_gpod and "Window_Delay_20Hz" in v:
                nat_mask = np.isnat(get_data_var(v))
                # convert window delay from ns to distance in m
                data_var = get_data_var(v).astype(float) * 1e-9 * CONST_C / 2
                data_var[nat_mask] = np.nan  #
            elif is_gpod and "Epoch_20Hz" in v:
                # wf_oversampling_factor = float(ds.WAVEFORM_OVERSAMPLING_FACTOR) if hasattr(ds, 'WAVEFORM_OVERSAMPLING_FACTOR') else 1
                # data_var = wf_oversampling_factor * get_data_var(v)  #
                # convert from s in ns
                data_var = get_data_var(v)  # convert from s in ns
            elif is_gpod and "time" in v and not is_cs:
                data_var = np.datetime64(ds.SENSING_START_UTC_TIME[:-1]) + get_data_var(
                    v
                )
            elif is_gpod and "time" in v and is_cs:
                dt_str = ds.SENSING_START[1:]
                dt_str = dt_str.replace(dt_str[3:6], dt_str[3:6].lower().title())
                data_var = np.datetime64(
                    datetime.strptime(dt_str, "%d-%b-%Y %H:%M:%S.%f")
                ) + get_data_var(v)
            elif is_gpod and "swh_qual" in k and "Misfit_20Hz" in v:
                # as defined in Dinardo2020, 3.4 (True=bad)
                data_var = get_data_var(v) > 4
            elif is_gpod and "Epoch_Reference_Gate" in v:
                # GPOD is 1-based, we need 0-based
                data_var = get_data_var(v) - 1
            # EUMETSAT or DeDop
            elif is_eumetsat and "lon" in k:
                # df.loc[df.lon >= 180.0, 'lon'] -= 360.0
                data_var = get_data_var(v)
                inds_to_convert = np.argwhere(get_data_var(v) > np.pi)
                data_var[inds_to_convert] -= 2 * np.pi
            elif is_eumetsat and "epoch_ref_gate_const" in v and is_s3:
                data_var = (
                    44.0 - 1
                )  # fixed, info from Bruno Lucas, Mail conversation from 6th of April
            elif is_eumetsat and "epoch_ref_gate_const" in v and is_s6:
                zp_factor = 2
                data_var = 128 * zp_factor
            elif (is_eumetsat) and "Vs" in k and is_s3:
                data_var = np.sqrt(
                    get_data_var("x_vel_l1b_echo_sar_ku") ** 2
                    + get_data_var("y_vel_l1b_echo_sar_ku") ** 2
                    + get_data_var("z_vel_l1b_echo_sar_ku") ** 2
                )
            elif (is_eumetsat or is_dart) and "Vs" in k and is_s6:
                v = get_data_var("velocity_vector")
                data_var = np.apply_along_axis(
                    lambda vel: np.sqrt(vel[0] ** 2 + vel[1] ** 2 + vel[2] ** 2), 1, v
                )
            elif is_dart and "time" in v:
                data_var = np.datetime64(
                    ds.first_measurement_time[:-1]
                ) + pd.to_timedelta(get_data_var(v), "s")
            # CS
            elif is_cs and "epoch_ref_gate_const" in v:
                data_var = 128.0
            elif is_cs and "beam_ang_stack" in k:
                angle_start = (
                    get_data_var("dop_angle_start_20_hr_ku")
                    if is_cs_gopr
                    else get_data_var("dop_angle_start_20_ku")
                )
                angle_stop = (
                    get_data_var("dop_angle_stop_20_hr_ku")
                    if is_cs_gopr
                    else get_data_var("dop_angle_stop_20_ku")
                )
                n_beams = (
                    get_data_var("stack_number_after_weighting_20_hr_ku").astype(int)
                    if is_cs_gopr
                    else get_data_var("stack_number_after_weighting_20_ku").astype(int)
                )

                def func(start, stop, num):
                    arr = np.full(200, np.nan)
                    arr[:num] = np.linspace(start, stop, num) - np.pi / 2
                    return arr

                data_var = np.array(
                    [
                        func(start, stop, num)
                        for start, stop, num in zip(angle_start, angle_stop, n_beams)
                    ]
                )
            elif is_cs and "Vs" in k:
                data_var = np.sqrt(
                    get_data_var(v)[0, 0] ** 2
                    + get_data_var(v)[0, 1] ** 2
                    + get_data_var(v)[0, 2] ** 2
                )
            elif is_cs and "ksi" in k:
                data_var = get_data_var(v)
            elif is_cs and "tracker" in k:
                # data_var = get_data_var(v).astype(float) * 1e9
                data_var = get_data_var(v).astype(float) * 1e-9 * CONST_C / 2
            elif is_cs and "epoch" in k:
                trk_range = (
                    get_data_var("window_del_20_hr_ku")
                    if is_cs_gopr
                    else get_data_var("window_del_20_ku")
                )
                tracker_range_m = trk_range.astype(float) * 1e-9 * CONST_C / 2
                range_m = (
                    get_data_var("range_ocean_20_hr_ku")
                    if is_cs_gopr
                    else get_data_var("range_ocean_20_ku")
                )
                range_res_m = (1 / 320e6) * CONST_C / 2
                data_var = (
                    (range_m - tracker_range_m) * range_res_m * (1 / 320e6) / 1e-9
                )
            elif is_cs and "Pu_W" in k or "n_iter" in k:
                data_var = np.full(n_inds, np.nan)
            # EUMETSAT
            elif is_eumetsat and "wf" in k and is_s6:
                # data_var = get_data_var(v) * get_data_var('waveform_scale_factor')
                data_var = get_data_var(v)
            elif is_dart and "beam_ang_stack" in k:
                data_var = get_data_var(v) + np.pi / 2
            elif is_eumetsat and "beam_ang_stack" in v and is_s3:
                data_var = get_data_var(v) + np.pi / 2
            elif is_eumetsat and "beam_ang_stack" in k and is_s6:
                n_looks_max = int(ds.dims["looks"])

                angle_start = get_data_var("look_angle_start")
                angle_stop = get_data_var("look_angle_stop")
                num_looks = np.min(
                    [
                        get_data_var("num_looks_start_stop"),
                        get_data_var("num_looks_multilooking"),
                    ],
                    axis=0,
                ).astype(int)
                angle_per_look_rad = (angle_stop - angle_start) / (num_looks - 1)

                def func(start, num, angle_per_look, n_looks_max):
                    arr = np.full(n_looks_max, np.nan)
                    arr[:num] = start + np.arange(num) * angle_per_look + np.pi / 2
                    return arr

                angs = np.array(
                    [
                        func(start, num, angle_per_look, n_looks_max)
                        for start, num, angle_per_look in zip(
                            angle_start, num_looks, angle_per_look_rad
                        )
                    ]
                )

                data_var = np.fliplr(angs)
            elif is_eumetsat and is_s6 and "wf" in k:
                data_var = np.multiply(
                    get_data_var(v),
                    get_data_var("waveform_scale_factor")[:, np.newaxis],
                )
            elif is_eumetsat and "mqe_ocean_20_ku" in v:
                data_var = convert_eum_mqe_to_misfit(get_data_var(v))
            elif is_eumetsat and "epoch_ocean_20_ku" in v:
                data_var = get_data_var(v) / 1e-9  # convert from s in ns
            elif is_eumetsat and "alt" in k and is_s3:
                # for L1BS files, subtract center-of-gravity to antenna
                # distance from altitude
                if "1BS" in ds.title:
                    cog = get_data_var("cog_cor_l1b_echo_sar_ku")
                    data_var = get_data_var(v) - cog
                else:
                    data_var = get_data_var(v)
            elif is_s6 and "ksi" in k:
                data_var = np.radians(get_data_var(v))
            # all
            elif is_eumetsat and "misfit" in k:
                mqe_var = [v for k, v in data_var_names.items() if "mqe" in v][0]
                data_var = convert_eum_mqe_to_misfit(get_data_var(mqe_var)) > 4
            else:
                data_var = get_data_var(v)

            dataset[k] = data_var

    if len(dataset) != len(data_var_names):
        missing = [d for d in list(data_var_names) if d not in list(dataset)]
        raise RuntimeError(
            "Not all fields were found in datasets. Missing: {}".format(
                ",".join(missing)
            )
        )

    # add RIP if available (first found occurrence taken)
    rip_var_list = [d for d in ds.variables if "rip" in d.lower()]
    if rip_var_list:
        dataset["rip"] = get_data_var(rip_var_list[0])

    # add entropy, pulse_peakiness etc.
    if "wf" in dataset:
        wf_norm = (dataset["wf"].T / np.max(dataset["wf"], axis=1).T).T
        dataset["entropy"] = np.apply_along_axis(
            lambda wf: -np.nansum(np.square(wf) * np.log2(np.square(wf))), 1, wf_norm
        )
        dataset["pulse_peakiness"] = np.nanmax(dataset["wf"], axis=1) / np.nansum(
            dataset["wf"], axis=1
        )

    # dist2coast
    dataset["dist2coast"] = get_dist_pacioos(
        latarr=np.degrees(dataset["lat_rad"]),
        lonarr=np.degrees(dataset["lon_rad"]),
        do_interp=do_interp_dist2coast,
    )
    # dirty hack: we assume land-water-transitions are set nan -> set them to
    # 0.0 so that we retrack them
    dataset["dist2coast"][np.isnan(dataset["dist2coast"])] = 0.0

    # add ascending or descending track boolean var
    if is_gpod and "PASS_DIRECTION_START" in ds.attrs:
        dataset["ascending"] = "ascending" == ds.PASS_DIRECTION_START
    elif dataset["lat_rad"].shape[0] > 1:
        dataset["ascending"] = bool(dataset["lat_rad"][1] > dataset["lat_rad"][0])

    if "record_ind" in ds:
        dataset["record_inds"] = ds.record_ind.values[n_offset : last_ind + 1]
    else:
        dataset["record_inds"] = np.arange(n_offset, last_ind + 1)

    return dataset


def get_model_param_obj_from_l1b_data(dataset: dict, ind: int):
    d = {
        k: (v if isinstance(v, (bool, float, int)) or len(v) <= 1 else v[ind])
        for k, v in dataset.items()
        if v is not None
    }
    return ModelParameter(**d)


def get_subset_dataset(dataset: dict, ind_offset: int, n_inds: int = 1):
    d = {}

    for k, v in dataset.items():
        if np.isscalar(v):
            d[k] = v
        elif isinstance(v, (np.ndarray, np.generic)):
            if v.size == 1:
                d[k] = v[0]
            elif v.size > 1 and len(v.shape) > 1:
                d[k] = v[ind_offset : ind_offset + n_inds, :]
            elif v.size > 1 and len(v.shape) == 1:
                d[k] = v[ind_offset : ind_offset + n_inds]

                d[k] = d[k][0] if d[k].size == 1 else d[k]
            else:
                RuntimeError("Invalid type")

            if len(d[k].shape) > 1 and d[k].shape[0] == 1:
                d[k] = d[k].squeeze()
        else:
            RuntimeError("Invalid type")

    return d


def gen_subset_dataset(dataset, rel_inds):
    for ind in rel_inds:
        yield get_subset_dataset(dataset, ind)


def gen_model_param_obj_from_l1b_data(dataset, rel_inds):
    for ind in rel_inds:
        yield get_model_param_obj_from_l1b_data(dataset, ind)


def get_nc_src_dest_file_list(l1b_source, nc_dest_dir):
    nc_dest_dir = nc_dest_dir if nc_dest_dir else Path(tempfile.gettempdir())
    abs_dest_dir = nc_dest_dir.resolve()

    if not isinstance(l1b_source, list) and os.path.isdir(l1b_source):
        src_files = list(l1b_source.rglob("/*.nc"))
        src_files_from_base = [f.parent for f in src_files]
        dest_files = [abs_dest_dir / f for f in src_files_from_base]
    elif not isinstance(l1b_source, list):
        src_files = list(l1b_source)

        # add further level directory if files without any unique identifier,
        # e.g. measurement_l1bs.nc
        src_files_basename = (
            src_files[0].parent.stem
            if "measurement" in src_files[0].stem and "SEN6" not in src_files[0].stem
            else src_files[0].stem
        )

        if "SEN" in src_files_basename and "measurement" in src_files_basename:
            dest_files = [
                abs_dest_dir / f'{f.name.split(".SEN6.")[0]}.nc' for f in src_files
            ]
        elif "S3" not in src_files_basename and "measurement" in src_files_basename:
            dest_files = []
            for f in src_files:
                track_id = f.parent.parent.name
                file_id = f.parent.name
                dest_files.append(
                    abs_dest_dir / track_id / file_id / src_files_basename
                )
        elif "SEN" in str(src_files[0].parent.name):
            dest_files = [abs_dest_dir / f.parent.name / f.name for f in src_files]
        elif "SEN" in src_files[0].parent.name:
            dest_files = [abs_dest_dir / f.parent.name / f.name for f in src_files]
        else:
            dest_files = [abs_dest_dir / f.name for f in src_files]
    else:
        src_files = l1b_source
        if "measurement" in str(src_files[0]):
            dest_files = [
                abs_dest_dir / (f.parent.name.replace(".SEN6", "") + ".nc")
                for f in src_files
            ]
        else:
            dest_files = [abs_dest_dir / f.name for f in src_files]

    return src_files, dest_files


def get_nc_dir_df(nc_files):
    all_nc_files = sorted(nc_files.rglob("*.nc"))
    is_l2 = "_l2_" in str(all_nc_files[0]).lower()

    def convert_dt(dt_str):
        return np.datetime64(datetime.strptime(dt_str, "%Y%m%dT%H%M%S"))

    if is_l2:
        cycles = [int(f.name.split("_")[8]) for f in all_nc_files]
        passes = [int(f.name.split("_")[9]) for f in all_nc_files]
        start_dates = [convert_dt(f.name.split("_")[10]) for f in all_nc_files]
        end_dates = [convert_dt(f.name.split("_")[11]) for f in all_nc_files]
    else:
        cycles = [int(f.name.split("_")[13]) for f in all_nc_files]
        passes = [int(f.name.split("_")[14]) for f in all_nc_files]
        start_dates = [convert_dt(f.name.split("_")[9]) for f in all_nc_files]
        end_dates = [convert_dt(f.name.split("_")[10]) for f in all_nc_files]

    df = pd.DataFrame(
        {
            "file": all_nc_files,
            "cycle": cycles,
            "ppass": passes,
            "start_date": start_dates,
            "end_date": end_dates,
        }
    )

    return df



import logging
from enum import Enum
from pathlib import Path

import numpy as np
from pydantic import BaseModel, root_validator, validator

from typing import Optional


class SensorType(Enum):
    S3 = "s3"
    CS = "cs"
    S6_F04 = "s6_f04"
    S6_F06 = "s6_f06"
    S6_F04_FF = "s6_f04_ff"
    S6_F06_FF = "s6_f06_ff"


class L1bSourceType(Enum):
    GPOD = "gpod"
    EUM_S3 = "eum_s3"
    EUM_CS = "eum_cs"
    EUM_S6_F04 = "eum_s6_f04"
    EUM_S6_F06 = "eum_s6_f06"
    EUM_S6_F04_FFSAR = "eum_s6_ffsar_f04"
    EUM_S6_F06_FFSAR = "eum_s6_ffsar_f06"


class RetrackerBaseType(Enum):
    SAM = "sam"
    SAMPLUS = "samplus"
    SAMPLUSPLUS = "samplusplus"


class SensorSettings(BaseModel):
    sensor_type: SensorType = SensorType.S3
    f_c_Hz: float = 13.575e9  # central frequency [Hz]
    B_t_Hz: float = 320e6  # chirp bandwidth [Hz]
    B_r_Hz: float = 320e6  # receiver bandwidth [Hz]
    # 3dB beamwidth along track [rad], from S.Dinardo (mail 29.05.20)
    theta_x_rad: float = np.radians(1.338)
    # 3dB beamwidth across track [rad], from S.Dinardo (mail 29.05.20)
    theta_y_rad: float = np.radians(1.338)
    # Pulse repetition frequency [Hz] (Determined by reverse engineering,
    # confirmed by W. Smith), from S.Dinardo (mail 02.06.20), 4488 = pulse
    # repetition interval
    prf_Hz: float = 1 / (4488 / 80e6)

    # tau_u: float = 44.8e-6  # useful pulse length [s]
    n_b: int = 64  # number of pulses per Burst [-]
    # Burst Repetition Interval [s] (Determined by reverse engineering,
    # confirmed by W. Smith)
    bri: float = 1018710 * 1 / 80e6

    sh_x: float = 1.0  # antenna shape factor along track [-]
    sh_y: float = 1.0  # antenna shape factor across track [-]

    def get_default_sets(st: SensorType):
        sensor_sets = SensorSettings()

        if st is SensorType.S3:
            pass
        elif st is SensorType.CS:
            sensor_sets.sensor_type = SensorType.CS
            sensor_sets.B_r_Hz = 320e6
            sensor_sets.theta_x_rad = np.radians(1.06)
            sensor_sets.theta_y_rad = np.radians(1.1992)
            sensor_sets.prf_Hz = 1 / (4400 * (1 / 80e6))
            sensor_sets.bri = None
        elif "s6" in st.value:
            sensor_sets.sensor_type = st
            sensor_sets.B_r_Hz = 395e6
            theta_x_y = 1.33
            sensor_sets.theta_x_rad = np.radians(theta_x_y)
            sensor_sets.theta_y_rad = np.radians(theta_x_y)
            sensor_sets.prf_Hz = None
            sensor_sets.bri = None

        return sensor_sets


SENSOR_SETS_DEFAULT_S3 = SensorSettings.get_default_sets(SensorType.S3)


class WaveformSettings(BaseModel):
    zp_oversampling_factor: int = None
    internal_oversampling_factor: int = 1
    np: Optional[int]

    def get_default_src_type(st: L1bSourceType, **kwargs):
        if st is L1bSourceType.GPOD:
            return WaveformSettings(
                **{**{"zp_oversampling_factor": 2, "np": 256, **kwargs}}
            )
        elif st is L1bSourceType.EUM_S3:
            return WaveformSettings(
                **{**{"zp_oversampling_factor": 1, "np": 128, **kwargs}}
            )
        elif st is L1bSourceType.EUM_CS:
            return WaveformSettings(
                **{**{"zp_oversampling_factor": 2, "np": 128, **kwargs}}
            )
        elif "s6" in str(st).lower():
            return WaveformSettings(
                **{**{"zp_oversampling_factor": 2, "np": 256, **kwargs}}
            )

    @validator("zp_oversampling_factor", pre=True, always=True)
    def set_default_zp_oversampling_factor(cls, v, *, values, **kwargs):
        return v or 1

    @validator("np", pre=True, always=True)
    def set_default_np(cls, v, *, values, **kwargs):
        np = 128 if v is None else v
        return (
            np
            * values["zp_oversampling_factor"]
            * values["internal_oversampling_factor"]
        )


class RetrackerProcessorSettings(BaseModel):
    retracker_basetype: RetrackerBaseType
    n_offset: int = 0
    n_inds: int = 5  # 0 = all
    n_procs: int = None  # None = all available
    nc_dest_dir: Path = Path().cwd().parent / ".testrun"

    do_interp_dist2coast: bool = False
    do_write_out_nc: bool = True
    do_write_out_log: bool = False
    do_create_settings_log_file: bool = True
    skip_if_exists: bool = True

    do_dynamic_fg_epoch: bool = False
    dynamic_fg_epoch_n_adjacent_meas: int = 20

    auto_detect_s6_rmc: bool = True

    reduce_l2_factor: int = 1


class RetrackerSettings(BaseModel):
    retracker_basetype: RetrackerBaseType
    # flag to disable the usage of multilook (0 -> Enable ML; 1 -> Disable
    # Multilook)
    Disable_ML_Flag: bool = False
    # Flag to control activation of waveform normalization (1 = true ; 0 =
    # false);
    Wf_Norm_Flag: bool = True
    # Wf_Norm_Aw: int = 1  # Width (in gates) of the sliding window for waveform normalization
    # Wf_Norm_First: int = 1  # First gate of the waveform to search for the max. amplitude
    # Wf_Norm_Last: int = None  # Last gate of the waveform to search for the
    # max. amplitude
    Wf_TN_First: int = 5  # First gate of the waveform to estimate the amplitude of the thermal noise floor, without zero-padding factor
    Wf_TN_Last: int = 10  # Last gate of the waveform to estimate the amplitude of the thermal noise floor, without zero-padding factor
    # -1 means fit from beginning of waveform, without zero-padding factor
    Wf_Fit_First_Bin: int = -1
    # -1 means fit until end of waveform, without zero-padding factor
    Wf_Fit_Last_Bin: int = -1
    # Wf_Fit_Activate_Flag = 1 #Flag to control the activation of the waveform
    # fitting (1 = true ; 0 = false);
    Thr: float = (
        3.0  # Threshold parameter greater than 1 used to test Wf_max against TN
    )
    # TN_Flag = 0 #flag to determine how to deal with Thermal Noise (TN_Flag=1
    # -> retracked ; TN_Flag=0 constant/estimated)

    second_retracking_step_samplus: bool = False
    n_effective_looks: int = 0
    # 0:disabled, >0:number of samples around dynamic_fg_epoch
    normalise_wf_by_fg_region: int = 0
    # increases the weights for the detected leading edge
    leading_edge_weight_factor: float = 1.0

    interference_masking: bool = False
    interference_masking_grow: int = 0
    interference_masking_swh_max: float = 18.0
    interference_masking_mask_before_le: bool = False
    interference_masking_second_retracking_step: bool = False

    # auto-detects RAW/RMC mode for S6 and adjusts
    # Wf_Fit_First_Bin/Wf_Fit_Last_Bin
    auto_detect_s6_rmc: bool = True

    subwaveform_mode: bool = False
    subwaveform_n_gates_after_le: int = 5

    # fits the zero-doppler beam (reasonable for FF-SAR-processed waveforms)
    fit_zero_doppler: bool = False

    def get_default_sets(st: SensorType, **kwargs):
        retrack_sets = RetrackerSettings(**kwargs)

        if st is SensorType.S3:
            pass
        elif st is SensorType.CS:
            retrack_sets.Wf_TN_First = 3  # 0-based
            retrack_sets.Wf_TN_Last = 8  # 0-based
        elif "s6" in st.value:
            retrack_sets.Wf_TN_First = 13  # 0-based
            retrack_sets.Wf_TN_Last = 17  # 0-based
            retrack_sets.Wf_Fit_First_Bin = 11  # 0-based
            retrack_sets.Wf_Fit_Last_Bin = 132  # 0-based

            if "ff" in st.value:
                retrack_sets.fit_zero_doppler = True

        return retrack_sets


class ModelParameter(BaseModel):
    lat_rad: float = np.radians(48.0)
    alt_m: float = 815000
    Vs_m_per_s: float = 7500
    h_rate_m_per_s: float = 0.06
    ascending: bool = True
    ksix_rad: float = 0.0
    ksiy_rad: float = 0.0
    beam_ang_stack_rad: np.ndarray = np.array([])
    epoch_ref_gate: float = 65.0
    dist2coast: float = 42.0
    rip: np.ndarray = np.array([])
    pri_hz: Optional[float] = None
    stack_mask_start_stop: Optional[np.ndarray] = None

    @root_validator(pre=True)
    def warning_defaults(cls, values):
        fields_not_set = [f for f in cls.__fields__ if f not in values]
        if fields_not_set:
            logging.debug(
                "WARNING: the following params were not set, now taking defaults: {}".format(
                    ",".join(fields_not_set)
                )
            )
        return values

    class Config:
        arbitrary_types_allowed = True


class ModelSettings(BaseModel):
    retracker_basetype: RetrackerBaseType
    # flag to disable the computation of first order term in the SAMOSA model
    # (0 -> Enable First Order term; 1 -> Disable First order term)
    Disable_SAM_F1_Flag: bool = False
    # flag to disable the computation of alph_P by LUT (0 -> Use LUT to
    # extract the alph_P value; 1 -> use a constant value for alph_P)
    Disable_LUT_Flag: bool = False
    # Index for the first idealized beam accumulated in the multilooking stage
    Ideal_First_Look_Index_ML: int = -106
    # Index for the last idealized beam accumulated in the multilooking stage
    Ideal_Last_Look_Index_ML: int = 106
    # 1 -> idealized beam angles are computed from simple geometric
    # considerations; 0 -> Use the values from data.
    Ideal_Beam_Ang_Stack_flag: bool = False
    # flag to enable the slope effect compensation (1 -> Enable Slope Effect
    # Compensation; 0 -> Disable Slope Effect Compensation)
    Enable_Slope_Effect_Flag: bool = False
    # as in Dinardo2020, not as in DPM 2.5.2: 1 / (0.886 * np.sqrt(2 * np.pi))
    # # ~0.45, PTR Gaussian approximation coefficient value
    alpha_p_mean: float = 0.5
    n_effective_looks: int = 0
    # fits the zero-doppler beam (reasonable for FF-SAR-processed waveforms)
    fit_zero_doppler: bool = False

    def get_default_sets(st: SensorType, **kwargs):
        model_sets = ModelSettings(**kwargs)

        if st is SensorType.S3:
            pass
        elif st is SensorType.CS:
            model_sets.alpha_p_mean = 1 / (0.886 * np.sqrt(2 * np.pi))
        elif "s6" in st.value:
            model_sets.alpha_p_mean = 0.55
            model_sets.Enable_Slope_Effect_Flag = False

        return model_sets


class FittingParameters(BaseModel):
    fit_nu_instead_of_swh: bool = False
    swh_first_step: float = None
    init_epoch: tuple = None
    init_swh: tuple = None
    init_pu: tuple = None
    init_nu: tuple = None
    minmax_epoch: tuple = None
    minmax_swh: tuple = None
    minmax_pu: tuple = None
    minmax_nu: tuple = None
    disable_interference_masking: bool = False


class FittingSettings(BaseModel):
    retracker_basetype: RetrackerBaseType
    # Initial guess for the 2nd fitted variable (Hs)
    Fit_Var_2_Init_Hs: float = 2.0
    # Initial guess for the 3rd fitted variable (Pu)
    Fit_Var_3_Init_Pu: float = 1.0
    # Initial guess for the 4th fitted variable (ThNoise). Activge if
    # TN_Flag=1.
    Fit_Var_4_Init_TN: float = 0.0
    # Initial guess for the optional fitted variable (nu).
    Fit_Var_5_Init_nu: float = 0
    # Minimum/maximum allowed value for the 1st fitted variable (t0) in
    # nanoseconds
    Fit_Var_1_MinMax_epoch: tuple = (-600.0, 600.0)
    # Minimum/maximum allowed value for the 2nd fitted variable (Hs)
    Fit_Var_2_MinMax_Hs: tuple = (-0.5, 20)
    # Minimum/maximum allowed value for the 3rd fitted variable (Pu)
    Fit_Var_3_MinMax_Pu: tuple = (0.2, 1.5)
    # Minimum/maximum allowed value for the 4th fitted variable (TN).
    Fit_Var_4_MinMax_TN: tuple = (0.0, 0.05)
    # Minimum/maximum allowed value for the optional fitted variable (nu)
    Fit_Var_5_MinMax_nu: tuple = (0, 1e9)
    Fit_MaxIter: int = 20  # Maximum Number of iterations for the fitting routine}
    Levmar_Control_0: float = 1e-3  # Scalar factor of Damping
    Levmar_Control_1: float = 1e-3  # Threshold for Infinite-Norm of Error Gradient
    Levmar_Control_2: float = 1e-2  # Threshold for 2-Norm of distance from solution
    Levmar_Control_3: float = 1e-6  # Threshold for 2-Norm of Error (RMSE)
    Levmar_Control_4: float = 1e-3  # Step used in finite differences for jacobian
    # Threshold for xtol, ftol, and gtol tolerances of TRF fitting algorithm
    trf_threshold: float = 1e-5
    # Threshold for xtol, ftol, and gtol tolerances of TRF fitting algorithm
    trf_stepsize: float = 1e-2
    # sets strict epoch boundaries in gates around fg epoch (0: disabled)
    limit_epoch_around_fg_n: int = 0

    def get_default_sets(st: SensorType, **kwargs):
        fitting_sets = FittingSettings(**kwargs)

        if st is SensorType.S3:
            pass
        elif st is SensorType.CS:
            fitting_sets.trf_stepsize = 1e-2
        elif "s6" in st.value:
            pass

        return fitting_sets


class ExportSettings(BaseModel):
    rp_sets: RetrackerProcessorSettings
    retrack_sets: RetrackerSettings
    fitting_sets: FittingSettings
    sensor_sets: SensorSettings


class ProcMode(Enum):
    FFSAR = "ffsar"
    UFSAR = "ufsar"



import logging
import tempfile
from pathlib import Path
from textwrap import wrap
import xarray as xr

import numpy as np
import matplotlib.pyplot as plt
import pytest

from pysamosa import retracker
from pysamosa import simple_logger
from pysamosa.common_types import (
    L1bSourceType,
    RetrackerBaseType,
    ModelSettings,
    ProcMode,
)
from pysamosa.data_access import (
    get_subset_dataset,
    get_model_param_obj_from_l1b_data,
    data_vars_dart,
    data_vars_retracker,
    data_vars_s6,
)
from pysamosa.retracker_helpers import get_dynamic_first_guess_epochs
from pysamosa.settings_manager import get_default_base_settings, SettingsPreset
from tests.helpers import plot_retrack_result
from tests.settings_dumper import SettingsDumper


rel_inds, file_id = (
    # list(range(46403, 46403+60)), 's3_0',  # nice coastal retracking
    # scenario from CORAL paper
    [25480],
    "s3_0",  # nice coastal retracking scenario from CORAL paper
    # [1000], 'cs_0',
    # [3503], 'cs_1',
    # [3503], 'cs_2',
    # [13344], 's6_eum_0_f03',
    # [18600], 's6_eum_1_f04',
    # [18600], 's6_eum_2_f06',
    #
    # [7], 'dart_s6_0',
    # list(range(55,60)), 'dart_s6_0',
)

conf = [
    (L1bSourceType.EUM_S3, RetrackerBaseType.SAM, SettingsPreset.NONE),
    # (L1bSourceType.EUM_S3, RetrackerBaseType.SAM, SettingsPreset.CORALv1),
    ##
    # CS
    # (L1bSourceType.EUM_CS, RetrackerBaseType.SAM, SettingsPreset.NONE),
    # (L1bSourceType.EUM_CS, RetrackerBaseType.SAMPLUS, SettingsPreset.NONE),
    # S6
    # (L1bSourceType.EUM_S6_F04, RetrackerBaseType.SAM, SettingsPreset.NONE),
    # S6 FFSAR
    # (L1bSourceType.EUM_S6_FFSAR, RetrackerBaseType.SAM, SettingsPreset.NONE),
    # (L1bSourceType.EUM_S6_FFSAR, RetrackerBaseType.SAM, SettingsPreset.CORALv2),
    # (L1bSourceType.EUM_S6_F04, RetrackerBaseType.SAM, SettingsPreset.NONE),
    # (L1bSourceType.EUM_S6_F06, RetrackerBaseType.SAM, SettingsPreset.NONE),
    # (L1bSourceType.EUM_S6_F04, RetrackerBaseType.SAMPLUS, SettingsPreset.NONE),
    # (L1bSourceType.EUM_S6_F06, RetrackerBaseType.SAM, SettingsPreset.CORALv2),
]
l1b_src_type, retracker_basetype, settings_preset = conf[0]

rp_sets, retrack_sets, fitting_sets, wf_sets, sensor_sets = get_default_base_settings(
    retracker_basetype=retracker_basetype,
    settings_preset=settings_preset,
    l1b_src_type=l1b_src_type,
)
model_sets = ModelSettings.get_default_sets(
    st=sensor_sets.sensor_type,
    retracker_basetype=rp_sets.retracker_basetype,
    n_effective_looks=retrack_sets.n_effective_looks,
    wf_sets=wf_sets,
)


custom_proc_mode = None
# custom_proc_mode = ProcMode.FFSAR
# custom_proc_mode = ProcMode.UFSAR

custom_file_l1b, custom_file_l2 = None, None
if custom_proc_mode is not None:
    custom_file_l1b = Path(
        "/nfs/DGFI145/C/work_flo/coastal_ffsar/orig_data/f06/P4_1B_HR_____/S6A_P4_1B_HR______20210502T223038_20210502T232547_20220505T002438_3309_017_196_098_EUM__REP_NT_F06.SEN6/measurement.nc"
    )
    custom_file_l2 = Path(
        "/nfs/DGFI145/C/work_flo/coastal_ffsar/orig_data/f06/P4_2__HR_____/S6A_P4_2__HR______20210413T182810_20210413T192223_20220514T174711_3253_015_213_106_EUM__REP_NT_F06.SEN6/S6A_P4_2__HR_STD__NT_015_213_20210413T182810_20210413T192223_F06.nc"
    )
    rel_inds, file_id = list(range(8580, 8595)), None

    if "_DAR_" in str(custom_file_l1b):
        custom_datavars_l1b = (
            data_vars_dart["l1b"]
            if custom_proc_mode is ProcMode.UFSAR
            else data_vars_dart["l1b_ffsar"]
        )
    else:
        custom_datavars_l1b = data_vars_s6["l1b"]

    with xr.open_dataset(custom_file_l2, decode_times=False) as ds:
        if "processor" in ds.attrs and "DGFI" in ds.processor:
            custom_datavars_l2 = data_vars_retracker["l2"]
        else:
            custom_datavars_l2 = data_vars_s6["l2"]

    retrack_sets.fit_zero_doppler = (
        custom_proc_mode and custom_proc_mode is ProcMode.FFSAR
    )


@pytest.fixture(scope="class")
def set_dumper():
    tempdir = Path(tempfile.gettempdir())
    dest_mat_file = (
        tempdir
        / f"retrack_log_{file_id}_inds{rel_inds[0]}-{rel_inds[-1]}_{l1b_src_type.value}_{retracker_basetype.value}_{settings_preset.value}.mat"
    )
    dest_gif_file = (
        tempdir
        / f"retrack_log_{file_id}_inds{rel_inds[0]}-{rel_inds[-1]}_{l1b_src_type.value}_{retracker_basetype.value}_{settings_preset.value}.gif"
    )
    settings_dump = SettingsDumper(
        wf_sets=wf_sets,
        model_sets=model_sets,
        sensor_sets=sensor_sets,
        dest_mat_file=dest_mat_file,
        dest_gif_file=dest_gif_file,
    )
    yield settings_dump
    settings_dump.write_out_mat()
    settings_dump.export_gif()


fig_list = []


class TestRetrackingSingle:
    @pytest.mark.parametrize("rel_ind", rel_inds)
    def test_retracking_single(
        self,
        set_dumper,
        rel_ind,
        s3_eum_l1b,
        s3_eum_l2,
        cs_eum_l1b,
        cs_eum_l2,
        s6_eum_l1b,
        s6_eum_l2,
        s6_dart_l1b,
        s6_dart_l2,
        dataset_generic_l1b,
        dataset_generic_l2,
    ):
        n_offset_l2 = None

        if custom_proc_mode:
            l1b, l2 = dataset_generic_l1b(
                custom_file_l1b, custom_datavars_l1b
            ), dataset_generic_l2(custom_file_l2, custom_datavars_l2)
        elif l1b_src_type == L1bSourceType.EUM_S3:
            l1b, l2 = s3_eum_l1b, s3_eum_l2
        elif l1b_src_type == L1bSourceType.EUM_CS:
            l1b, l2 = cs_eum_l1b, cs_eum_l2
        elif (
            l1b_src_type == L1bSourceType.EUM_S6_F04
            or l1b_src_type == L1bSourceType.EUM_S6_F06
        ):
            l1b, l2 = s6_eum_l1b, s6_eum_l2
        elif (
            l1b_src_type == L1bSourceType.EUM_S6_F04_FFSAR
            or l1b_src_type == L1bSourceType.EUM_S6_F06_FFSAR
        ):
            l1b, l2 = s6_dart_l1b, s6_dart_l2
            retrack_sets.fit_zero_doppler = True

        ncfile_l1bs = l1b.get_nc_filename(file_id)
        if set_dumper and set_dumper.l1b_nc_filename is None:
            set_dumper.l1b_nc_filename = ncfile_l1bs

        n_offset = rel_ind

        # analyse and plot data
        fig, ax = plt.subplots()

        # calculate dynamic firt-guess epoch for a single measurement
        if rp_sets.do_dynamic_fg_epoch:
            n_inds_total = rp_sets.dynamic_fg_epoch_n_adjacent_meas
            n_before = n_inds_total // 2

            n_offset_fg = 0 if (n_offset - n_before) < 0 else n_offset - n_before
            n_target_ind_l1b = n_offset if n_offset_fg == 0 else n_before
            l1b_data = l1b(n_offset=n_offset_fg, n_inds=n_inds_total, file_id=file_id)
            l1b_data["dynamic_fg_epoch"] = get_dynamic_first_guess_epochs(
                wfs=l1b_data["wf"],
                tracker_range=l1b_data["tracker_range_m"],
                alt_m=l1b_data["alt_m"],
                bu_bw_Hz=sensor_sets.B_r_Hz,
                fg_epoch_adjacent_meas=rp_sets.dynamic_fg_epoch_n_adjacent_meas,
            )
            l1b_data_single = get_subset_dataset(l1b_data, ind_offset=n_target_ind_l1b)
        else:
            n_target_ind_l1b = 0
            # choose two to get right direction of track
            l1b_data = l1b(n_offset=n_offset, n_inds=2, file_id=file_id)
            l1b_data_single = get_subset_dataset(l1b_data, ind_offset=n_target_ind_l1b)

        try:
            l2_record_inds_all = l2(n_offset=0, n_inds=0, file_id=file_id)[
                "record_inds"
            ]
            n_ind_l2 = np.argwhere(
                l1b_data_single["record_inds"] == l2_record_inds_all
            )[0][0]
        except BaseException:
            n_ind_l2 = n_offset_l2 if (n_offset_l2 is None) else n_offset

        l2_data = l2(n_offset=n_ind_l2, n_inds=1, file_id=file_id)
        l2_data_single = get_subset_dataset(l2_data, ind_offset=0)
        model_params = get_model_param_obj_from_l1b_data(l1b_data, ind=n_target_ind_l1b)

        # simple_logger.set_root_logger()
        simple_logger.set_root_logger(log_level=logging.DEBUG)

        sr = retracker.SamosaRetracker(
            retracker_basetype=retracker_basetype,
            retrack_sets=retrack_sets,
            fitting_sets=fitting_sets,
            sensor_sets=sensor_sets,
            wf_sets=wf_sets,
        )

        # start fitting
        res_fit = sr.fit_wf(l1b_data_single=l1b_data_single, model_params=model_params)

        # plot fitted curve
        plot_retrack_result(
            l1b_data_single,
            l2_data_single,
            res_fit,
            model_params=model_params,
            wf_sets=wf_sets,
            model_sets=sr.model_sets,
            sensor_sets=sensor_sets,
            retrack_sets=retrack_sets,
            ax=ax,
            show_l2_model=False,
            show_second_halve=l1b_src_type == L1bSourceType.GPOD,
        )

        fn = ncfile_l1bs.name if ncfile_l1bs is not None else ""
        fig.suptitle(
            "\n".join(
                wrap(
                    f"{fn}, {retracker_basetype.value}-{settings_preset.value} ({l1b_src_type.value}), record#: {n_ind_l2}",
                    width=70,
                )
            ),
            fontsize=8,
        )
        # fig.subplots_adjust(top=0.65)
        fig.subplots_adjust(top=0.77)
        fig.show()

        if set_dumper:
            set_dumper.add_retrack_entry(
                l1b_data_single=l1b_data_single,
                model_params=model_params,
                record_ind=n_ind_l2,
                res_fit=res_fit,
                l2_data_single=l2_data_single,
            )
            set_dumper.add_fig(fig)



from enum import Enum

from pysamosa.common_types import (
    RetrackerProcessorSettings,
    RetrackerSettings,
    FittingSettings,
    RetrackerBaseType,
    WaveformSettings,
    SensorType,
    SensorSettings,
    L1bSourceType,
)


class SettingsPreset(Enum):
    NONE = "none"  # take default preset for RetrackerBaseType
    SAM_FLO = "samflo"
    CORALv1 = "coralv1"
    CORALv2 = "coralv2"
    SAMPLUS_FLO = "samplusflo"


def get_default_base_settings(
    *,
    retracker_basetype: RetrackerBaseType,
    settings_preset: SettingsPreset = SettingsPreset.NONE,
    l1b_src_type: L1bSourceType
):
    if "s3" in l1b_src_type.value:
        sensor_type = SensorType.S3
    elif "cs" in l1b_src_type.value:
        sensor_type = SensorType.CS
    elif "s6" in l1b_src_type.value and "ff" in l1b_src_type.value:
        sensor_type = (
            SensorType.S6_F06_FF
            if "f06" in l1b_src_type.value
            else SensorType.S6_F04_FF
        )
    elif "s6" in l1b_src_type.value and "ff" not in l1b_src_type.value:
        sensor_type = (
            SensorType.S6_F06 if "f06" in l1b_src_type.value else SensorType.S6_F04
        )
    else:
        raise RuntimeError("error setting SensorType. ")

    sensor_sets = SensorSettings.get_default_sets(sensor_type)

    rp_sets = RetrackerProcessorSettings(retracker_basetype=retracker_basetype)
    retrack_sets = RetrackerSettings.get_default_sets(
        st=sensor_type, retracker_basetype=retracker_basetype
    )
    fitting_sets = FittingSettings.get_default_sets(
        st=sensor_type, retracker_basetype=retracker_basetype
    )

    wf_sets = WaveformSettings.get_default_src_type(l1b_src_type)

    if retracker_basetype == RetrackerBaseType.SAM:
        rp_sets.do_dynamic_fg_epoch = False
        # rp_sets.dynamic_fg_epoch_n_adjacent_meas = 30  # default: 20

        # retrack_sets.second_retracking_step_samplus = True
        # retrack_sets.n_effective_looks = 0
        # retrack_sets.normalise_wf_by_fg_region = 5
        # retrack_sets.leading_edge_weight_factor = 4.0
        # retrack_sets.interference_masking = True
        # retrack_sets.interference_masking_grow = 2
        # retrack_sets.interference_masking_swh_max = 8.0
        # retrack_sets.interference_masking_second_retracking_step = True

        # fitting_sets.Levmar_Control_1 = 1e-3
        # fitting_sets.Levmar_Control_2 = 1e-1
        # fitting_sets.Levmar_Control_3 = 1e-3
        # fitting_sets.Levmar_Control_4 = 1e-1
        # fitting_sets.lock_epoch_around_fg_n = True
    elif retracker_basetype == RetrackerBaseType.SAMPLUS:
        rp_sets.do_dynamic_fg_epoch = False
        rp_sets.dynamic_fg_epoch_n_adjacent_meas = 30  # default: 20

        retrack_sets.second_retracking_step_samplus = True
        # retrack_sets.n_effective_looks = 0
        # retrack_sets.normalise_wf_by_fg_region = 5
        # retrack_sets.leading_edge_weight_factor = 1.0
        # retrack_sets.interference_masking = True
        # retrack_sets.interference_masking_grow = 0
        # retrack_sets.interference_masking_swh_max = 8.0
        # retrack_sets.interference_masking_second_retracking_step = True

        # fitting_sets.Levmar_Control_2 = 1e-3
        # fitting_sets.Levmar_Control_4 = 1e-4
        # fitting_sets.lock_epoch_around_fg_n = True
        fitting_sets.trf_threshold = 1e-5
        fitting_sets.trf_stepsize = 1e-2

    elif retracker_basetype == RetrackerBaseType.SAMPLUSPLUS:
        # rp_sets.do_dynamic_fg_epoch = True
        # rp_sets.dynamic_fg_epoch_n_adjacent_meas = 20  # default: 20

        retrack_sets.second_retracking_step_samplus = False
        # retrack_sets.n_effective_looks = 0
        # retrack_sets.normalise_wf_by_fg_region = 5
        # retrack_sets.leading_edge_weight_factor = 1.0
        # retrack_sets.interference_masking = True
        # retrack_sets.interference_masking_grow = 0
        # retrack_sets.interference_masking_swh_max = 8.0
        # retrack_sets.interference_masking_second_retracking_step = Truewf_sets is not None

        # fitting_sets.Levmar_Control_2 = 1e-1
        # fitting_sets.lock_epoch_around_fg_n = True

    # PRESETS
    if (
        settings_preset == SettingsPreset.CORALv1
        or settings_preset == SettingsPreset.CORALv2
    ):
        rp_sets.do_dynamic_fg_epoch = True
        rp_sets.dynamic_fg_epoch_n_adjacent_meas = 40

        retrack_sets.second_retracking_step_samplus = True
        # retrack_sets.n_effective_looks = 0
        retrack_sets.normalise_wf_by_fg_region = 5
        # retrack_sets.leading_edge_weight_factor = 4.0
        retrack_sets.interference_masking = True
        retrack_sets.interference_masking_grow = 5
        retrack_sets.interference_masking_swh_max = 8.0
        retrack_sets.interference_masking_mask_before_le = False
        retrack_sets.interference_masking_second_retracking_step = True
        # retrack_sets.interference_retrack_after_cleansing = True

        # fitting_sets.Levmar_Control_2 = 1e-2
        fitting_sets.limit_epoch_around_fg_n = 10
        fitting_sets.trf_threshold = 1e-4
        fitting_sets.trf_stepsize = 1e-2

        if settings_preset == SettingsPreset.CORALv2:
            retrack_sets.interference_masking_mask_before_le = True
            fitting_sets.Fit_Var_2_MinMax_Hs = (0.0, 20)

    # adapt parameters according to internal_oversampling_factor
    if wf_sets.internal_oversampling_factor > 1 or wf_sets.zp_oversampling_factor > 1:
        mult_factor = (
            wf_sets.internal_oversampling_factor
            if wf_sets.zp_oversampling_factor == 1
            else wf_sets.zp_oversampling_factor
        )

        retrack_sets.normalise_wf_by_fg_region *= mult_factor
        retrack_sets.interference_masking_grow *= mult_factor
        fitting_sets.limit_epoch_around_fg_n *= mult_factor

    return rp_sets, retrack_sets, fitting_sets, wf_sets, sensor_sets



import logging

from pysamosa.common_types import RetrackerBaseType, L1bSourceType
from pysamosa.data_access import data_vars_s3
from pysamosa.retracker_processor import RetrackerProcessor
from pysamosa.settings_manager import get_default_base_settings, SettingsPreset
from pysamosa.settings import LFSDATA_DIR


if __name__ == "__main__":
    nc_src_base_path = LFSDATA_DIR / "s3" / "l1b"
    run_name = "s3_retrack_open_ocean"
    nc_dest_path = nc_src_base_path.parent

    # select files
    l1b_files = [f for f in sorted(nc_src_base_path.rglob("*.nc"))]

    l1b_src_type = L1bSourceType.EUM_S3
    # rbt = RetrackerBaseType.SAMPLUS
    rbt = RetrackerBaseType.SAM
    # pres = SettingsPreset.CORALv1
    pres = SettingsPreset.NONE
    (
        rp_sets,
        retrack_sets,
        fitting_sets,
        wf_sets,
        sensor_sets,
    ) = get_default_base_settings(
        retracker_basetype=rbt, settings_preset=pres, l1b_src_type=l1b_src_type
    )

    rp_sets.nc_dest_dir = nc_dest_path / run_name
    rp_sets.n_offset = 25800
    rp_sets.n_inds = 100
    rp_sets.n_procs = 6
    rp_sets.skip_if_exists = False

    additional_nc_attrs = {
        "L1B source type": l1b_src_type.value.upper(),
        "Retracker basetype": rbt.value.upper(),
        "Retracker preset": pres.value.upper(),
    }

    rp = RetrackerProcessor(
        l1b_source=l1b_files,
        l1b_data_vars=data_vars_s3["l1b"],
        rp_sets=rp_sets,
        retrack_sets=retrack_sets,
        fitting_sets=fitting_sets,
        wf_sets=wf_sets,
        sensor_sets=sensor_sets,
        nc_attrs_kw=additional_nc_attrs,
        log_level=logging.DEBUG,  # comment in to show debug messages
    )
    rp.process()



import matplotlib.pyplot as plt

from pysamosa.common_types import RetrackerBaseType, L1bSourceType
from pysamosa.settings_manager import get_default_base_settings, SettingsPreset

test_inds_gpod = [
    # coast
    # 12327,
    # 12330,
    # 12330,
    # 12330,
    # *[v for v in range(12330, 12340)],
    # 12339,
    # 12340,
    # 12341,
    # *[v for v in range(12340, 12360)],
    # 12343,
    # open-ocean
    # *list(range(41658, 41668)),
    42354,
]


# rp_sets, retrack_sets, fitting_sets = get_default_base_settings(RetrackerBaseType.SAMPLUS, SettingsPreset.NONE)
rp_sets, retrack_sets, _, _, _ = get_default_base_settings(
    retracker_basetype=RetrackerBaseType.SAMPLUS,
    settings_preset=SettingsPreset.SAM_FLO,
    l1b_src_type=L1bSourceType.EUM_S3,
)

# rp_sets.do_dynamic_fg_epoch = True
rp_sets.dynamic_fg_epoch_n_adjacent_meas = 30  # default: 20

# retrack_sets.n_effective_looks = 0
# retrack_sets.normalise_wf_by_fg_region = 5
# retrack_sets.interference_masking = True
# retrack_sets.interference_masking_grow = 0
# retrack_sets.interference_masking_swh_max = 8.0
# retrack_sets.second_retracking_step_samplus = True

# fitting_sets.Levmar_Control_2 = 1e-1
# fitting_sets.lock_epoch_around_fg_n = True


def plot_leading_edge_analysis(le_inds, l1b_data_single, ind_gpod, fg_epoch):
    # plotting
    fig, axs = plt.subplots(1, 1)
    axs.plot(l1b_data_single["wf"].T)

    # plot initial first-guess epoch
    fg_epoch_style = {
        "linestyle": "--",
        "linewidth": 1,
        "color": "grey",
        "label": "fg_epoch",
    }
    axs.axvline(l1b_data_single["dynamic_fg_epoch"], **fg_epoch_style)

    # plot leading edge area
    axs.axvspan(le_inds[0], le_inds[-1], alpha=0.5, color="orange")
    axs.grid()
    axs.set_title(
        f"ind_gpod={ind_gpod}, len(le_inds)={len(le_inds)}, fg_epoch={fg_epoch}",
        fontsize=6,
    )

    fig.show()



import numpy as np
from pysamosa.utils import consecutive_regions_from_ind_list


def test_consecutive_regions_from_ind_list():
    a = [
        53,
        54,
        55,
        56,
        57,
        58,
        59,
        60,
        61,
        62,
        63,
        64,
        65,
        66,
        67,
        68,
        69,
        70,
        71,
        72,
        73,
        74,
        75,
        76,
        77,
        78,
        79,
        80,
        81,
        82,
        83,
        84,
        85,
        86,
        87,
        88,
        89,
        90,
        91,
        92,
        93,
        94,
        95,
        96,
        97,
        98,
    ]

    regions = consecutive_regions_from_ind_list(a)
    assert len(regions) == 1
    assert np.array_equal(a, regions[0])

    a = [2, 3, 4, 6, 7]
    regions = consecutive_regions_from_ind_list(a)
    assert len(regions) == 2
    assert np.array_equal(regions[0], [2, 3, 4])
    assert np.array_equal(regions[1], [6, 7])

    a = []
    regions = consecutive_regions_from_ind_list(a)
    assert len(regions) == 0



from pysamosa.common_types import (
    RetrackerBaseType,
    FittingSettings,
    WaveformSettings,
    SensorType,
    L1bSourceType,
    ModelSettings,
)


def test_fitting_sets():
    fit_sets = FittingSettings(retracker_basetype=RetrackerBaseType.SAM)
    assert fit_sets.Fit_Var_2_Init_Hs == 2.0


def test_waveform_sets():
    wf_sets = WaveformSettings()
    assert wf_sets.zp_oversampling_factor == 1
    assert wf_sets.np == 128

    wf_sets = WaveformSettings(np=256, zp_oversampling_factor=2)
    assert wf_sets.zp_oversampling_factor == 2
    assert wf_sets.np == 512

    wf_sets = WaveformSettings(np=256, zp_oversampling_factor=2)
    assert wf_sets.zp_oversampling_factor == 2
    assert wf_sets.np == 512

    wf_sets = WaveformSettings(internal_oversampling_factor=4)
    assert wf_sets.np == 512

    wf_sets = WaveformSettings(internal_oversampling_factor=4, zp_oversampling_factor=2)
    assert wf_sets.np == 1024

    # defaults
    wf_sets = WaveformSettings.get_default_src_type(L1bSourceType.GPOD)
    assert wf_sets.np == 512
    assert wf_sets.zp_oversampling_factor == 2

    wf_sets = WaveformSettings.get_default_src_type(L1bSourceType.EUM_S3)
    assert wf_sets.np == 128
    assert wf_sets.zp_oversampling_factor == 1

    wf_sets = WaveformSettings.get_default_src_type(L1bSourceType.EUM_S6_F04)
    assert wf_sets.np == 512
    assert wf_sets.zp_oversampling_factor == 2

    wf_sets = WaveformSettings.get_default_src_type(L1bSourceType.EUM_S6_F06)
    assert wf_sets.np == 512
    assert wf_sets.zp_oversampling_factor == 2

    wf_sets = WaveformSettings.get_default_src_type(L1bSourceType.EUM_S6_F04_FFSAR)
    assert wf_sets.np == 512
    assert wf_sets.zp_oversampling_factor == 2

    wf_sets = WaveformSettings.get_default_src_type(L1bSourceType.EUM_S6_F06_FFSAR)
    assert wf_sets.np == 512
    assert wf_sets.zp_oversampling_factor == 2

    wf_sets = WaveformSettings.get_default_src_type(
        L1bSourceType.EUM_S6_F04, internal_oversampling_factor=2
    )
    assert wf_sets.np == 1024
    assert wf_sets.zp_oversampling_factor == 2

    wf_sets = WaveformSettings.get_default_src_type(
        L1bSourceType.EUM_S3, zp_oversampling_factor=2
    )
    assert wf_sets.zp_oversampling_factor == 2
    assert wf_sets.np == 256


def test_model_sets():
    ms = ModelSettings.get_default_sets(
        retracker_basetype=RetrackerBaseType.SAM, st=SensorType.S3
    )
    assert ms.alpha_p_mean == 0.5

    ms = ModelSettings.get_default_sets(
        retracker_basetype=RetrackerBaseType.SAM, st=SensorType.S6_F04
    )
    assert ms.alpha_p_mean == 0.55



from collections import deque
from itertools import islice
from concurrent import futures
from itertools import repeat
import numpy as np

from pysamosa.data_access import (
    _read_dataset_vars_from_ds,
    get_subset_dataset,
    get_model_param_obj_from_l1b_data,
)
from pysamosa.retracker import SamosaRetracker
from pysamosa.conf_params import CONST_C


def get_tracker_shift_n_samps(
    tracker_range_m: np.ndarray, alt_m: np.ndarray, t0_sampling_gate_ns: float
) -> np.ndarray:
    """
    Calulates the number of samples to shift for range-aligning multiple waveforms.

    :param tracker_range_m: distance between altimeter and the receiving window
    :param alt_m: altitude of altimeter
    :param t0_sampling_gate_ns: sampling time of one gate sample in the receiving window
    :return: an np.ndarray containing the integer number of samples to shift
    """
    t_diff = np.subtract(alt_m, tracker_range_m) / CONST_C
    t_corr_no_bias = np.subtract(t_diff, np.nanmean(t_diff))

    n_shift_samps = np.round(np.divide(t_corr_no_bias, t0_sampling_gate_ns))
    n_shift_samps = np.nan_to_num(n_shift_samps)  # set nan values to a 0-shift

    return n_shift_samps.astype(int)


def custom_roll(arr, r_tup):
    m = np.asarray(r_tup)
    # need `copy`
    arr_roll = arr[:, [*range(arr.shape[1]), *range(arr.shape[1] - 1)]].copy()
    strd_0, strd_1 = arr_roll.strides
    n = arr.shape[1]
    result = np.lib.stride_tricks.as_strided(
        arr_roll, (*arr.shape, n), (strd_0, strd_1, strd_1)
    )

    return result[np.arange(arr.shape[0]), (n - m) % n]


def shift_waveform_by_n(wfs: np.ndarray, shifts: np.ndarray):
    dims_orig = wfs.shape
    if wfs.shape[0] != len(shifts):
        raise RuntimeError(
            "shifts vector must be of the same length as the number of waveforms. "
        )

    wfs = wfs.copy()

    wfs = np.pad(wfs, ((0, 0), (0, np.max(shifts))), mode="constant")
    res = custom_roll(wfs, shifts)[:, : dims_orig[1]]
    return res


def get_pointwise_product(wfs: np.ndarray, axis=0) -> np.ndarray:
    op = "ij,ij->j" if axis == 0 else "ij,ij->i"
    return np.einsum(op, wfs, wfs)


def get_inds_sliding_window(n_total_len, n):
    """Generates a sliding window with the absolute indices.

    :param n_total_len: length of total sequence lengths
    :param n: sliding window size
    :return:
    """

    def sliding_window(seq, n=2):
        "Returns a sliding window (of width n) over data from the iterable"
        "   s -> (s0,s1,...s[n-1]), (s1,s2,...,sn), ...                   "
        it = iter(seq)
        result = list(islice(it, n))
        if len(result) == n:
            yield (result, n // 2)
        for elem in it:
            result = result[1:] + [
                elem,
            ]
            yield (result, n // 2)

    # pre- and append for n-1 times the same window to account for sliding
    # in/out effect
    if n_total_len > n:
        n_prepend = int(n // 2)
        n_append = n - (n_prepend + 1)
        windows_no_slidein = list(sliding_window(np.arange(n_total_len), n))
        windows = [
            *[(w, i) for i, w in enumerate([windows_no_slidein[0][0]] * n_prepend)],
            *windows_no_slidein,
            *[
                (w, (n // 2 + 1 + i))
                for i, w in enumerate([windows_no_slidein[-1][0]] * n_append)
            ],
        ]
    else:
        windows = [
            (win, i) for i, win in enumerate([list(range(n_total_len))] * n_total_len)
        ]

    return windows


def gen_sliding_window_from_seq(seq, n):
    """
    Generates a sliding window over an iterable sequence.

    :param seq: the iterable sequence the sliding window is generated from
    :param n: number of surrounding samples, the total sliding window size is n+1, n must be even-numbered
    :return: the sliding window of the current iteration
    """
    if n % 2:
        raise ValueError("n must be even-numbered. ")
    if len(seq) < n:
        raise ValueError("len(seq) must be larger than n. ")

    center_sample = 0

    half_window = n // 2
    n_total = n + 1
    it = iter(seq)
    win = deque((next(it, None) for _ in range(n + 1)), maxlen=n_total)

    # sliding window in
    for i in range(half_window + 1):
        yield win, i
    center_sample = half_window

    # sliding inside
    for e in it:
        win.append(e)
        if center_sample < n // 2:
            center_sample += 1
        yield win, center_sample

    # sliding window out
    for i in range(center_sample + 1, n_total):
        yield win, i


def get_single_dynamic_first_guess_epochs(
    wfs: np.ndarray,
    tracker_range: np.ndarray,
    alt_m: np.ndarray,
    bu_bw_Hz,
    shifted_window_abs_inds_center: list,
):
    n_gates = wfs.shape[1]

    # shifted_window = shifted_window_abs_inds_center[0]
    rel_ind_window_center = shifted_window_abs_inds_center[1]

    # get shift in n samples that aligns all waveforms in terms of its
    # absolute ranges
    n_shifts = get_tracker_shift_n_samps(
        tracker_range_m=tracker_range, alt_m=alt_m, t0_sampling_gate_ns=1 / bu_bw_Hz
    )

    # if there is a very abrupt jump in the tracker range, then limit the
    # adjacent samples to have a maximum shift of half of the wf length
    mask_possible_shifts = np.abs((n_shifts - n_shifts[rel_ind_window_center])) < (
        n_gates // 3
    )

    # adjust possible indices
    rel_ind_window_center_possible = rel_ind_window_center - np.sum(
        ~mask_possible_shifts[0:rel_ind_window_center]
    )  # accounts for invalid tracker range jumps
    n_shifts = n_shifts[mask_possible_shifts]
    n_shifts -= int(np.mean(n_shifts))  # demean

    wfs_range_aligned = shift_waveform_by_n(wfs[mask_possible_shifts, :], -n_shifts)

    # wf_current_shifted = wfs_range_aligned[rel_ind_window_center_possible]

    # normalised range-aligned wf
    wfs_range_aligned_norm = np.apply_along_axis(
        lambda wf: wf / np.max(wf), 1, wfs_range_aligned
    )

    if all(
        np.apply_along_axis(
            lambda wf: np.allclose(wf[~np.isnan(wf)], 0.0),
            axis=1,
            arr=wfs_range_aligned_norm,
        )
    ):
        return 0, np.full((n_gates,), np.nan), np.full((n_gates,), np.nan)

    # if there are wfs with all-zero or all-nan, set them to 1 -> no effect on
    # pointwise product
    wfs_range_aligned_norm = np.apply_along_axis(
        lambda wf: np.ones(len(wf))
        if all(np.isnan(wf)) or np.isclose(np.mean(wf), 0.0, atol=1e-25)
        else wf,
        1,
        wfs_range_aligned_norm,
    )
    wfs_pw_prod = get_pointwise_product(wfs_range_aligned_norm)

    ind_max_prod = np.argmax(wfs_pw_prod)
    ind_max_adjacent_wfs = ind_max_prod

    new_fg_epoch = ind_max_adjacent_wfs + n_shifts[rel_ind_window_center_possible]

    return new_fg_epoch, wfs_range_aligned_norm, wfs_pw_prod


def get_dynamic_first_guess_epochs(
    wfs: np.ndarray,
    tracker_range: np.ndarray,
    alt_m: np.ndarray,
    bu_bw_Hz,
    fg_epoch_adjacent_meas=20,
    n_procs=1,
) -> np.ndarray:
    """Calculates the dynamic first-guess epochs based on the pointwise product of the n_adjacent_meas waveforms

    e.g. for n_adjacent_meas=20, for calculating the new fg epoch for waveform W_n:
    - account for alt_m_n - tracker_range_n -> shift W_n correspondingly
    - take  point-wise product of these tracker_range_n-aligned waveforms: W_n_-10, W_n_-9, ..., W_n, ..., W_n_9

    :param wfs: np.ndarray containing the waveforms (n_adjacent_samps, n_window_length)
    :param tracker_range: range of tracking window (L1B parameter)
    :param alt_m: altitude of satellite (L1B parameter)
    :param bu_bw_Hz: receiver bandwidth (L1B parameter)
    :param fg_epoch_adjacent_meas: number of adjacent measurements to take into account
    :param n_procs: number of processes, if None all available logical cores will be used
    :return: new first-guess epochs in gates, starting from 0
    """
    new_fg_epochs = np.ones(wfs.shape[0], dtype=int)
    n_gates = wfs.shape[1]

    if n_procs == 1:
        for abs_ind, shifted_window_abs_inds_center in enumerate(
            get_inds_sliding_window(wfs.shape[0], n=fg_epoch_adjacent_meas)
        ):
            wfs_window = wfs[shifted_window_abs_inds_center[0],]
            tracker_range_window = tracker_range[shifted_window_abs_inds_center[0]]
            alt_m_window = alt_m[shifted_window_abs_inds_center[0]]

            new_fg_epoch, _, _ = get_single_dynamic_first_guess_epochs(
                wfs=wfs_window,
                tracker_range=tracker_range_window,
                alt_m=alt_m_window,
                bu_bw_Hz=bu_bw_Hz,
                shifted_window_abs_inds_center=shifted_window_abs_inds_center,
            )

            if 0 < new_fg_epoch < n_gates:
                new_fg_epochs[abs_ind] = new_fg_epoch
            else:
                new_fg_epochs[abs_ind] = np.argmax(wfs[abs_ind])
    else:
        with futures.ProcessPoolExecutor(max_workers=n_procs) as pool:
            shifted_window_abs_inds_center = get_inds_sliding_window(
                wfs.shape[0], n=fg_epoch_adjacent_meas
            )

            def gen_subset_view(arr, shifted_window_abs_inds_center):
                for i in range(len(shifted_window_abs_inds_center)):
                    yield arr[shifted_window_abs_inds_center[i][0],]

            wfs_gen = gen_subset_view(wfs, shifted_window_abs_inds_center)
            alt_m_gen = gen_subset_view(alt_m, shifted_window_abs_inds_center)
            tracker_range_gen = gen_subset_view(
                tracker_range, shifted_window_abs_inds_center
            )

            for abs_ind, (new_fg_epoch, _, _) in enumerate(
                pool.map(
                    get_single_dynamic_first_guess_epochs,
                    wfs_gen,
                    tracker_range_gen,
                    alt_m_gen,
                    repeat(bu_bw_Hz),
                    shifted_window_abs_inds_center,
                    chunksize=100,
                )
            ):
                if 0 < new_fg_epoch < n_gates:
                    new_fg_epochs[abs_ind] = new_fg_epoch
                else:
                    new_fg_epochs[abs_ind] = np.argmax(wfs[abs_ind])

    return new_fg_epochs


def retrack_single_meas(*, sr, rp_sets, n_offset, nc_l1bs, datavars, sensor_sets):
    # calculate dynamic firt-guess epoch for a single measurement
    if rp_sets.do_dynamic_fg_epoch:
        # read n samples before the selected ones, required for estimation
        # dynamic first guess epoch
        n_inds_total = rp_sets.dynamic_fg_epoch_n_adjacent_meas
        n_pre = n_inds_total // 2

        l1b_data = _read_dataset_vars_from_ds(
            nc_filename=nc_l1bs,
            data_var_names=datavars,
            n_offset=n_offset - n_pre,
            n_inds=n_inds_total,
        )
        l1b_data["dynamic_fg_epoch"] = get_dynamic_first_guess_epochs(
            wfs=l1b_data["wf"],
            tracker_range=l1b_data["tracker_range_m"],
            alt_m=l1b_data["alt_m"],
            bu_bw_Hz=sensor_sets.B_r_Hz,
            fg_epoch_adjacent_meas=rp_sets.dynamic_fg_epoch_n_adjacent_meas,
        )
        l1b_data_single = get_subset_dataset(l1b_data, ind_offset=n_pre)
        model_params = get_model_param_obj_from_l1b_data(l1b_data, ind=n_pre)
    else:
        l1b_data = _read_dataset_vars_from_ds(
            nc_filename=nc_l1bs, data_var_names=datavars, n_offset=n_offset, n_inds=1
        )
        l1b_data_single = get_subset_dataset(l1b_data, ind_offset=0)
        model_params = get_model_param_obj_from_l1b_data(l1b_data, ind=0)

    # start fitting
    res_fit = sr.fit_wf(l1b_data_single=l1b_data_single, model_params=model_params)

    fg_epoch = (
        l1b_data_single["dynamic_fg_epoch"]
        if "dynamic_fg_epoch" in l1b_data_single
        else np.argmax(l1b_data_single["wf"])
    )
    fg_epoch = fg_epoch * sr.wf_sets.internal_oversampling_factor

    if sr.retrack_sets and sr.retrack_sets.normalise_wf_by_fg_region:
        wf_meas = res_fit["wf"] / SamosaRetracker.get_wf_max(
            res_fit["wf"], retrack_sets=sr.retrack_sets, fg_epoch=fg_epoch
        )
    else:
        wf_meas = res_fit["wf"] / np.max(res_fit["wf"])

    return wf_meas, res_fit, l1b_data_single, model_params, fg_epoch



import numpy as np
from scipy.signal import argrelextrema
from scipy.interpolate import interp1d
from pydantic import BaseModel
import copy

from pysamosa.conf_params import CONST_C
from pysamosa.common_types import ModelParameter, SensorSettings

from astropy import modeling
from astropy.modeling import models


# class RipAnalyserSettings(BaseModel):
#     model_params: ModelParameter
#     sensor_sets: SensorSettings


class RIPParameters(BaseModel):
    rip_az_fitted: np.ndarray = None
    nu: float = None
    mss: float = None
    pitch_mispoint_look: float = None
    pitch_mispoint_rad: float = None
    amplitude_fitted_norm: float = None
    fresnel_coeff_rf0: float = None
    halfpower_aperture_rad: float = None
    halfpower_looks: tuple = None

    class Config:
        arbitrary_types_allowed = True


def get_n_negative_looks(n_looks):
    return n_looks // 2 if n_looks % 2 else n_looks // 2 - 1


class RipAnalyser:
    def __init__(
        self,
        rip_wf: np.ndarray,
        sensor_sets: SensorSettings,
        model_params: ModelParameter,
    ):
        self.rip_wf = copy.deepcopy(rip_wf).reshape(
            [
                -1,
            ]
        )
        self.sensor_sets = sensor_sets
        self.model_params = model_params

        # oversample RIP before doing the fitting
        self.oversampling_factor = 8

        # normalize RIP
        self.rip_max = np.max(self.rip_wf)
        self.rip_wf_norm = self.rip_wf / self.rip_max

        self._rip_wf_len = self.rip_wf_norm.size
        self._n_negative_looks = get_n_negative_looks(self._rip_wf_len)
        self.doppler_beam_inds = np.arange(
            -self._n_negative_looks, self._rip_wf_len // 2 + 1
        )

        # calculate sensor-dependent params
        self.gamma_x = (
            self.sensor_sets.sh_x * 8 * np.log(2) / (self.sensor_sets.theta_x_rad**2)
        )
        Tb = self.sensor_sets.n_b / self.sensor_sets.prf_Hz
        self.Lx = (CONST_C * self.model_params.alt_m) / (
            2 * self.model_params.Vs_m_per_s * self.sensor_sets.f_c_Hz * Tb
        )
        self.dtheta_rad = self.Lx / self.model_params.alt_m
        self.theta_looks_rad = self.dtheta_rad * self.doppler_beam_inds

        # fit RIP
        self.rip_params = self._fit_along_track_rip_wf()

    def _fit_along_track_rip_wf(self):
        if self.oversampling_factor > 1.0:
            # oversample RIP
            f = interp1d(self.doppler_beam_inds, self.rip_wf_norm, kind="cubic")

            doppler_beam_inds = np.linspace(
                self.doppler_beam_inds[0],
                self.doppler_beam_inds[-1],
                num=self.oversampling_factor * self._rip_wf_len,
            )
            theta_looks_rad = np.linspace(
                self.theta_looks_rad[0],
                self.theta_looks_rad[-1],
                num=self.oversampling_factor * self._rip_wf_len,
            )
            self.rip_wf_oversampled = f(doppler_beam_inds)
        else:
            doppler_beam_inds = self.doppler_beam_inds
            theta_looks_rad = self.theta_looks_rad
            self.rip_wf_oversampled = self.rip_wf_norm

        # fit along-track RP using astropy.modeling.fitting module
        fitter = modeling.fitting.LevMarLSQFitter()
        model = models.Gaussian1D(amplitude=1.0, mean=0, stddev=1.0)
        self.rip_fitted_model = fitter(
            model, doppler_beam_inds, self.rip_wf_oversampled
        )

        # calc RIP params from fitted gaussian
        nu = (
            1 / (2 * (self.rip_fitted_model.stddev * self.dtheta_rad) ** 2)
            - self.gamma_x
        )
        mss = np.sqrt(1 / nu)
        fresnel_coeff_rf0 = mss * np.sqrt(2 * self.rip_max)

        self.rip_az_fitted = self.rip_fitted_model(doppler_beam_inds)

        # half-power aperture (depending on sea state, rough surface -> diffuse
        # RIP -> larger 3db aperture)
        n_inds_hpbw = argrelextrema(
            np.abs(self.rip_az_fitted - 0.5 * self.rip_fitted_model.amplitude.value),
            np.less,
        )[0]

        # check if gaussian fitted worked as expected
        if len(n_inds_hpbw) < 2:
            raise RuntimeError("RIP-Gaussian fitting failed. ")

        n_looks_hpbw = n_inds_hpbw[1] - n_inds_hpbw[0]
        hp_aperture_rad = n_looks_hpbw * self.dtheta_rad

        self.rip_params = RIPParameters(
            rip_az_fitted=self.rip_az_fitted[:: self.oversampling_factor],
            nu=nu,
            mss=mss,
            pitch_mispoint_look=self.rip_fitted_model.mean.value,
            pitch_mispoint_rad=np.interp(
                x=self.rip_fitted_model.mean.value,
                xp=doppler_beam_inds,
                fp=theta_looks_rad,
            ),
            amplitude_fitted_norm=self.rip_fitted_model.amplitude.value,
            fresnel_coeff_rf0=fresnel_coeff_rf0,
            halfpower_aperture_rad=hp_aperture_rad,
            halfpower_looks=tuple(
                (n_inds_hpbw / self.oversampling_factor - self._n_negative_looks)
            ),
        )

        return self.rip_params

    def get_gamma0(self, n_looks_eff, n_gates):
        """Calculates the gamma0 matrix/2D RIP. Interpolates the across-track RIP_act (which cannot be measured) from the along-track RIP_az

        :param n_looks_eff: number of effective looks that is used for the retracker (when generating the multilooked SAMOSA waveform model)
        :param n_gates: number of gates of the SAR power echo waveform
        :return: GAMMA0 matrix with shape (Np, Neff)
        """
        n_negative_looks_eff = get_n_negative_looks(n_looks_eff)
        self.looks_eff = np.arange(
            -n_negative_looks_eff, (n_looks_eff - n_negative_looks_eff)
        )

        # interpolate across-track RIP by taking the inner n_looks_eff effective looks
        # self.looks_rip_act = np.arange(0, (n_looks_eff - n_negative_looks_eff))
        self.looks_rip_act = np.linspace(
            0, (n_looks_eff - n_negative_looks_eff) - 1, num=n_gates
        )
        rip_act = self.rip_fitted_model(
            self.looks_rip_act + self.rip_fitted_model.mean.value
        )
        # f = interp1d(self.looks_rip_act, rip_act, kind='cubic')

        # self.gates = np.linspace(0, self.looks_rip_act[-1], num=n_gates)
        # self.rip_act_eff = f(self.gates)
        self.rip_act_eff = rip_act

        # calculate 2d RIP (along- and across-track)
        wf_start_look = get_n_negative_looks(self.rip_wf.size) - n_negative_looks_eff
        inds_act_eff = np.arange(wf_start_look, wf_start_look + n_looks_eff)
        self.gamma0_2d_rip = np.outer(
            self.rip_act_eff,
            self.rip_max
            * self.rip_az_fitted[:: self.oversampling_factor][inds_act_eff],
        )

        # normalise
        self.gamma0_2d_rip = self.gamma0_2d_rip / np.max(self.gamma0_2d_rip)

        return self.gamma0_2d_rip



import logging
import os
from concurrent import futures
from itertools import repeat

import numpy as np
import xarray as xr
import netCDF4

from pysamosa import simple_logger
from pysamosa.common_types import (
    RetrackerProcessorSettings,
    RetrackerSettings,
    FittingSettings,
    WaveformSettings,
    ExportSettings,
    L1bSourceType,
    SensorSettings,
)
from pysamosa.conf_params import CONST_C
from pysamosa.data_access import (
    _read_dataset_vars_from_ds,
    get_subset_dataset,
    get_model_param_obj_from_l1b_data,
    get_nc_src_dest_file_list,
    gen_subset_dataset,
    gen_model_param_obj_from_l1b_data,
)
from pysamosa.retracker import SamosaRetracker
from pysamosa.retracker_helpers import get_dynamic_first_guess_epochs
from pysamosa.version import __version__


def haversine_np(lat1, lon1, lat2, lon2):
    """
    Calculate the great circle distance between two points
    on the earth (specified in decimal degrees)

    All args must be of equal length.

    """
    lon1, lat1, lon2, lat2 = map(np.radians, [lon1, lat1, lon2, lat2])

    dlon = lon2 - lon1
    dlat = lat2 - lat1

    a = np.sin(dlat / 2.0) ** 2 + np.cos(lat1) * np.cos(lat2) * np.sin(dlon / 2.0) ** 2

    c = 2 * np.arcsin(np.sqrt(a))
    dist_km = 6378 * c
    return dist_km


class RetrackerProcessor:
    def __init__(
        self,
        *,
        l1b_source,
        l1b_data_vars,
        bbox=None,
        rp_sets: RetrackerProcessorSettings,
        retrack_sets: RetrackerSettings,
        fitting_sets: FittingSettings,
        sensor_sets: SensorSettings,
        wf_sets: WaveformSettings,
        nc_attrs_kw: dict = None,
        log_level: int = logging.INFO,
    ):
        simple_logger.set_root_logger(log_level=log_level)

        self.rt = rp_sets.retracker_basetype

        self.l1b_data_vars = l1b_data_vars
        self.rp_sets = (
            RetrackerProcessorSettings(retracker_basetype=self.rt)
            if rp_sets is None
            else rp_sets
        )
        self.l1b_src_files, self.l2_dest_files = get_nc_src_dest_file_list(
            l1b_source, self.rp_sets.nc_dest_dir
        )
        self.bbox = (
            bbox if bbox is not None else [None for i in range(len(self.l1b_src_files))]
        )

        logging.info(
            f"Started retracking processing of {len(self.l1b_src_files)} L1B files!"
            + f"(n_cores={os.cpu_count() if rp_sets.n_procs is None else rp_sets.n_procs})\n"
            + (
                ", ".join([f"{k}: {v}" for k, v in nc_attrs_kw.items()])
                if nc_attrs_kw is not None
                else ""
            )
        )

        self.retrack_sets = (
            RetrackerSettings(retracker_basetype=self.rt)
            if retrack_sets is None
            else retrack_sets
        )
        self.fitting_sets = (
            FittingSettings(retracker_basetype=self.rt)
            if fitting_sets is None
            else fitting_sets
        )
        self.sensor_sets = sensor_sets if sensor_sets is not None else SensorSettings()
        self.wf_sets = (
            wf_sets
            if wf_sets is not None
            else WaveformSettings.get_default_src_type(L1bSourceType.EUM_S3)
        )
        self.nc_attrs_kw = nc_attrs_kw if nc_attrs_kw is not None else {}

        self.retracker = SamosaRetracker(
            retracker_basetype=self.rt,
            retrack_sets=self.retrack_sets,
            fitting_sets=self.fitting_sets,
            sensor_sets=self.sensor_sets,
            wf_sets=self.wf_sets,
        )

        if self.rp_sets.do_create_settings_log_file:
            self.write_settings_log_file()

    def get_init_output_l2(self, *, reduce_factor=1, reduce_offset=0):
        if reduce_factor > 1:
            unique_vals, inds = np.unique(
                self.l1b_data["record_inds"], return_index=True
            )
            inds += reduce_offset
        else:
            # assumes that we process all entries in l1b_data
            inds = np.arange(len(self.l1b_data["record_inds"]))
        n_inds = len(inds)

        # pre-allocate output Dataset
        return xr.Dataset.from_dict(
            {
                "data_vars": {
                    "swh": {
                        "data": np.full(n_inds, np.nan),
                        "dims": "time",
                        "attrs": {"units": "m"},
                    },
                    "swh_qual": {
                        "data": np.ones(n_inds),
                        "dims": "time",
                        "attrs": {
                            "units": "m",
                            "long_name": "(conservative) quality flag for the swh variable",
                            "flag_values": "[0 1]",
                            "flag_meaning": "good bad",
                        },
                    },
                    "swh_alt_qual": {
                        "data": np.ones(n_inds),
                        "dims": "time",
                        "attrs": {
                            "units": "m",
                            "long_name": "(alternative) quality flag for the swh variable",
                            "flag_values": "[0 1]",
                            "flag_meaning": "good bad",
                        },
                    },
                    "swh_alt2_qual": {
                        "data": np.ones(n_inds),
                        "dims": "time",
                        "attrs": {
                            "units": "m",
                            "long_name": "second (alternative) quality flag for the swh variable",
                            "flag_values": "[0 1]",
                            "flag_meaning": "good bad",
                        },
                    },
                    "epoch": {
                        "data": np.full(n_inds, np.nan),
                        "dims": "time",
                        "attrs": {
                            "units": "ns",
                            "long_name": "epoch relative from epoch reference gate",
                        },
                    },
                    "range": {
                        "data": np.full(n_inds, np.nan),
                        "dims": "time",
                        "attrs": {
                            "units": "m",
                            "long_name": "retracked range at 20 Hz (no geo-corrections applied)",
                        },
                    },
                    "altitude": {
                        "data": self.l1b_data["alt_m"][inds],
                        "dims": "time",
                        "attrs": {"units": "m", "long_name": "altitude of satellite"},
                    },
                    "tracker_range": {
                        "data": self.l1b_data["tracker_range_m"][inds],
                        "dims": "time",
                        "attrs": {
                            "units": "m",
                            "long_name": "range of tracker window related to the epoch reference gate (epoch_ref_gate). ",
                        },
                    },
                    "Pu": {
                        "data": np.full(n_inds, np.nan),
                        "dims": "time",
                        "attrs": {"units": "W"},
                    },
                    "misfit": {
                        "data": np.full(n_inds, np.nan),
                        "dims": "time",
                        "attrs": {
                            "long_name": "misfit between fitted and fitted model waveform according to formula $misfit = 100 * sqrt(1/N * sum(wf - wf_model)**2)$"
                        },
                    },
                    "misfit_le": {
                        "data": np.full(n_inds, np.nan),
                        "dims": "time",
                        "attrs": {
                            "long_name": "leading-edge misfit between fitted and fitted model waveform's leading edge according to formula $misfit = 100 * sqrt(1/N * sum(wf - wf_model)**2)$"
                        },
                    },
                    "misfit_selective": {
                        "data": np.full(n_inds, np.nan),
                        "dims": "time",
                        "attrs": {
                            "long_name": "selective (interference is excluded) misfit between fitted and fitted model waveform according to formula $misfit = 100 * sqrt(1/N * sum(wf - wf_model)**2)$"
                        },
                    },
                    "dist2coast": {
                        "data": self.l1b_data["dist2coast"][inds],
                        "dims": "time",
                        "attrs": {"units": "km"},
                    },
                    "record_ind": {
                        "data": self.l1b_data["record_inds"][inds],
                        "dims": "time",
                        "attrs": {
                            "units": "count",
                            "long_name": "absolute record index as taken from L1BS file",
                        },
                    },
                },
                "coords": {
                    "time": {"data": self.l1b_data["time"][inds], "dims": "time"},
                    "latitude": {
                        "data": np.degrees(self.l1b_data["lat_rad"][inds]),
                        "dims": "latitude",
                        "attrs": {"units": "degree"},
                    },
                    "longitude": {
                        "data": np.degrees(self.l1b_data["lon_rad"][inds]),
                        "dims": "longitude",
                        "attrs": {"units": "degree"},
                    },
                },
                "attrs": {
                    **{
                        "processor": "PySAMOSA L2 Retracker for Delay-Doppler Altimetry",
                        "processor version": __version__,
                        "epoch_ref_gate": int(self.l1b_data["epoch_ref_gate"]),
                    },
                    **self.nc_attrs_kw,
                },
            }
        )

    @classmethod
    def _process_single(cls, retracker_obj, l1b_data_single, model_params, abs_ind):
        logging.debug(f"Processing waveform #{abs_ind}")
        try:
            res_fit = retracker_obj.fit_wf(
                l1b_data_single=l1b_data_single, model_params=model_params
            )
        except (ValueError, RuntimeError) as e:
            logging.error(
                f'Error: Stopped retracking of waveform #{abs_ind}: {" ".join(e.args)}'
            )
            return {}

        logging.debug(
            f"Finished waveform #{abs_ind} (n_iter={res_fit['n_iter']}). SWH={res_fit['swh']:.2f}m,"
            f"epoch={res_fit['epoch_ns']:.2f}ns, Pu={res_fit['Pu']:.2f}W. "
        )

        return res_fit

    def _write_fit_results(self, rel_ind, res_fit):
        if res_fit:
            self.output_l2.swh[rel_ind] = res_fit["swh"]
            self.output_l2.swh_qual[rel_ind] = res_fit["misfit_selective"] > 4
            self.output_l2.swh_alt_qual[rel_ind] = res_fit["misfit_le"] > 4
            # as defined in Dinardo2020, 3.4 (True=bad)
            self.output_l2.swh_alt2_qual[rel_ind] = res_fit["misfit"] > 4

            self.output_l2.epoch[rel_ind] = res_fit["epoch_ns"]
            self.output_l2.range[rel_ind] = self.l1b_data["tracker_range_m"][
                rel_ind
            ] + (float(res_fit["epoch_ns"])) * 1e-9 * (CONST_C / 2)
            self.output_l2.Pu[rel_ind] = res_fit["Pu"]
            self.output_l2.misfit[rel_ind] = res_fit["misfit"]
            self.output_l2.misfit_le[rel_ind] = res_fit["misfit_le"]
            self.output_l2.misfit_selective[rel_ind] = res_fit["misfit_selective"]

    def process(self):
        for nc_src_file, nc_dest_filepath, bbox in zip(
            self.l1b_src_files, self.l2_dest_files, self.bbox
        ):
            try:
                nc_dest_dir = nc_dest_filepath.parent
                nc_base_id = (
                    nc_dest_filepath.parent.stem
                    if "measurement" in nc_dest_filepath.stem
                    else nc_dest_filepath.stem
                )

                if (
                    self.rp_sets.do_write_out_nc
                    and self.rp_sets.skip_if_exists
                    and nc_dest_filepath.exists()
                ):
                    logging.info(f"{nc_base_id} already exists, skipping...")
                    continue

                log_filename = nc_dest_dir / f"log_{nc_base_id}.txt"
                if self.rp_sets.do_write_out_log:
                    file_logger_handle = simple_logger.generate_and_add_file_logger(
                        log_filename
                    )

                logging.info(f"Reading of L1B dataset {nc_base_id}... ")
                grps = [
                    g for g in list(netCDF4.Dataset(nc_src_file).groups) if "data" in g
                ]
                nc_grp = f"{grps[0]}/ku" if grps else ""

                # if bbox is not None replace n_offset and n_inds by bbox logic
                if bbox is not None:
                    # hack required because some nc files have issues with
                    # decoding of times
                    try:
                        ds = xr.open_dataset(nc_src_file, group=nc_grp)
                    except BaseException:
                        ds = xr.open_dataset(
                            nc_src_file, group=nc_grp, decode_times=False
                        )

                    lat_deg = ds[self.l1b_data_vars["lat_rad"]].values
                    lon_deg = ds[self.l1b_data_vars["lon_rad"]].values

                    ds.close()

                    eps = 0.001
                    inds_in_bbox = sorted(
                        np.where(
                            (lat_deg > (bbox[0] - eps))
                            & (lat_deg < (bbox[1] + eps))
                            & (lon_deg > (bbox[2] - eps))
                            & (lon_deg < (bbox[3] + eps))
                        )[0]
                    )
                    self.rp_sets.n_offset = int(inds_in_bbox[0])
                    self.rp_sets.n_inds = len(inds_in_bbox)

                self.l1b_data = _read_dataset_vars_from_ds(
                    nc_filename=nc_src_file,
                    data_var_names=self.l1b_data_vars,
                    n_offset=self.rp_sets.n_offset,
                    n_inds=self.rp_sets.n_inds,
                    group=nc_grp,
                )

                logging.info("Finished reading.")

                if self.rp_sets.do_dynamic_fg_epoch:
                    logging.info("Calculating dynamic_first_guess_epochs... ")

                    fg_epoch = get_dynamic_first_guess_epochs(
                        wfs=self.l1b_data["wf"],
                        tracker_range=self.l1b_data["tracker_range_m"],
                        alt_m=self.l1b_data["alt_m"],
                        bu_bw_Hz=self.sensor_sets.B_r_Hz,
                        fg_epoch_adjacent_meas=self.rp_sets.dynamic_fg_epoch_n_adjacent_meas,
                        n_procs=self.rp_sets.n_procs,
                    )
                    self.l1b_data["dynamic_fg_epoch"] = fg_epoch

                    logging.info("Finished calculating dynamic_first_guess_epochs.")

                self.output_l2 = self.get_init_output_l2()

                # select indices to process: ocean waveforms only
                mask_inds_process = (self.output_l2.dist2coast >= 0).values
                abs_inds_process = (
                    self.rp_sets.n_offset + np.arange(self.output_l2.swh.shape[0])
                )[mask_inds_process]
                rel_inds_process = mask_inds_process.nonzero()[0]

                # Start processing
                logging.info(f"Started processing of L1B dataset {nc_base_id}... ")
                if self.rp_sets.n_procs == 1:
                    for rel_ind, abs_ind in zip(rel_inds_process, abs_inds_process):
                        res_fit = RetrackerProcessor._process_single(
                            retracker_obj=self.retracker,
                            l1b_data_single=get_subset_dataset(self.l1b_data, rel_ind),
                            model_params=get_model_param_obj_from_l1b_data(
                                self.l1b_data, rel_ind
                            ),
                            abs_ind=abs_ind,
                        )
                        self._write_fit_results(rel_ind, res_fit)
                else:
                    with futures.ProcessPoolExecutor(
                        max_workers=self.rp_sets.n_procs
                    ) as pool:
                        for rel_ind, res_fit in zip(
                            rel_inds_process,
                            pool.map(
                                RetrackerProcessor._process_single,
                                repeat(self.retracker),
                                gen_subset_dataset(self.l1b_data, rel_inds_process),
                                gen_model_param_obj_from_l1b_data(
                                    self.l1b_data, rel_inds_process
                                ),
                                abs_inds_process,
                                chunksize=20,
                            ),
                        ):
                            self._write_fit_results(rel_ind, res_fit)

                logging.info(f"Finished processing of L1B dataset {nc_base_id}. ")

                # Write netCDF file
                if self.rp_sets.do_write_out_nc:
                    os.umask(0)
                    nc_dest_dir.mkdir(parents=True, exist_ok=True)
                    self.reduce_l2()
                    self.output_l2.to_netcdf(nc_dest_filepath)

                    logging.info(f"netCDF file written. {nc_dest_filepath}")
            except RuntimeError as e:
                logging.error(
                    f"Error has occurred during processing of file {nc_src_file}: {e}"
                )
                # logging.root.removeHandler(file_logger_handle)
                logging.shutdown()

        logging.info(
            f"Processing done. ({len(self.l1b_src_files)} netCDF files. RetrackerProcessorSettings{self.rp_sets})"
        )

    def reduce_l2(self):
        if self.rp_sets.reduce_l2_factor > 1:
            reduce_offset = (
                self.rp_sets.reduce_l2_factor // 2
                if (self.rp_sets.reduce_l2_factor % 2) != 0
                else self.rp_sets.reduce_l2_factor // 2 - 1
            )
            ds_reduced = self.get_init_output_l2(
                reduce_factor=self.rp_sets.reduce_l2_factor, reduce_offset=reduce_offset
            )
            not self.retrack_sets.fit_zero_doppler
            len(self.output_l2.record_ind)

            for rel_ind, rec_ind in enumerate(np.unique(self.output_l2.record_ind)):
                mask_sel = self.output_l2.record_ind == rec_ind
                abs_mask_good_recs = mask_sel & ~(self.output_l2.swh_qual == 1)

                if all(~abs_mask_good_recs):  # if all are bad estimates
                    avg_func = np.nanmean
                    ds_reduced.swh[rel_ind] = avg_func(self.output_l2.swh[mask_sel])
                    ds_reduced.swh_qual[rel_ind] = 1.0
                    ds_reduced.epoch[rel_ind] = avg_func(self.output_l2.epoch[mask_sel])
                    ds_reduced.range[rel_ind] = avg_func(self.output_l2.range[mask_sel])
                    ds_reduced.Pu[rel_ind] = avg_func(self.output_l2.Pu[mask_sel])
                else:
                    avg_func = np.nanmean
                    ds_reduced.swh[rel_ind] = avg_func(
                        self.output_l2.swh[abs_mask_good_recs]
                    )
                    ds_reduced.swh_qual[rel_ind] = (
                        float(self.output_l2.swh_qual[abs_mask_good_recs])
                        if np.isscalar(self.output_l2.swh_qual[abs_mask_good_recs])
                        else 0.0
                    )
                    ds_reduced.epoch[rel_ind] = avg_func(
                        self.output_l2.epoch[abs_mask_good_recs]
                    )
                    ds_reduced.range[rel_ind] = avg_func(
                        self.output_l2.range[abs_mask_good_recs]
                    )
                    ds_reduced.Pu[rel_ind] = avg_func(
                        self.output_l2.Pu[abs_mask_good_recs]
                    )

            self.output_l2 = ds_reduced

            logging.info(
                f"processed L2 dataset reduced by factor {self.rp_sets.reduce_l2_factor}. "
            )

    def write_settings_log_file(self):
        json_params = {"indent": 4, "separators": (",", ": ")}

        all_sets = ExportSettings(
            rp_sets=self.rp_sets,
            retrack_sets=self.retrack_sets,
            fitting_sets=self.fitting_sets,
            sensor_sets=self.sensor_sets,
        )

        dest_filepath_json = self.rp_sets.nc_dest_dir / "settings.json"
        os.umask(0)
        dest_filepath_json.parent.mkdir(parents=True, exist_ok=True)

        with open(dest_filepath_json, "w") as f:
            f.write(all_sets.json(**json_params))



import numpy as np
import xarray as xr
from pathlib import Path

ncfile_l1b = Path(
    "/nfs/DGFI145/C/work_flo/cs2_files_samplus_test/CS2_open_ocean/CS_LTA__SIR_SAR_1B_20150503T160800_20150503T161329_D001.nc"
)
ncfile_l2gop = Path(
    "/nfs/DGFI145/C/work_flo/cs2_files_samplus_test/CS2_open_ocean/CS_LTA__SIR_GOP_2__20150503T155002_20150503T163939_C001_0269.nc"
)

l1b = xr.open_dataset(ncfile_l1b)
l2 = xr.open_dataset(ncfile_l2gop)

mask_latminmax = (l2.lat_20_ku >= (l1b.lat_20_ku.min() - 0.0001)) & (
    l2.lat_20_ku <= l1b.lat_20_ku.max()
)
l2_realigned = {
    k: v[mask_latminmax].values for k, v in dict(l2).items() if "time_20_ku" in v.dims
}

for k, v in l2_realigned.items():
    try:
        new_da = xr.DataArray(
            np.array(v.data),
            dims=["time_20_ku"],
            coords=dict(
                time_20_ku=l1b.coords["time_20_ku"].values,
                lat_20_ku=(["time_20_ku"], l1b.coords["lat_20_ku"].values),
                lon_20_ku=(["time_20_ku"], l1b.coords["lon_20_ku"].values),
            ),
        )
    except BaseException:
        pass
    l2_realigned[k] = new_da

l2_realigned["window_del_20_ku"] = xr.DataArray(
    l1b.window_del_20_ku.values,
    dims=["time_20_ku"],
    coords=dict(
        time_20_ku=l1b.coords["time_20_ku"].values,
        lat_20_ku=(["time_20_ku"], l1b.coords["lat_20_ku"].values),
        lon_20_ku=(["time_20_ku"], l1b.coords["lon_20_ku"].values),
    ),
)

ds_l2_realigned = xr.Dataset(l2_realigned)

ncfile_l2_dest = ncfile_l1b.parent / (
    str(ncfile_l1b.stem).replace("1B", "2_") + "_realigned" + ".nc"
)
ds_l2_realigned.to_netcdf(ncfile_l2_dest)
# mask_latminmax = l1b.lat
